#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun  1 14:20:52 2021

@author: peruzzetto
"""

import matplotlib.pyplot as plt
import numpy as np

import os
import importlib
import warnings

import tilupy.notations as notations
import tilupy.plot as plt_fn
import tilupy.raster

RAW_STATES = ["hvert", "h", "ux", "uy"]

TEMPORAL_DATA_0D = ["ek", "vol"]
TEMPORAL_DATA_1D = [""]
TEMPORAL_DATA_2D = ["hvert", "h", "u", "ux", "uy", "hu", "hu2"]
STATIC_DATA_0D = []
STATIC_DATA_1D = []
STATIC_DATA_2D = []

TOPO_DATA_2D = ["z", "zinit", "costh"]

NP_OPERATORS = ["max", "mean", "std", "sum", "min"]
OTHER_OPERATORS = ["final", "init", "int"]
TIME_OPERATORS = ["final", "init", "int"]

COMPUTED_STATIC_DATA_2D = []
for stat in NP_OPERATORS + OTHER_OPERATORS:
    for name in TEMPORAL_DATA_2D:
        COMPUTED_STATIC_DATA_2D.append(name + "_" + stat)
STATIC_DATA_2D += COMPUTED_STATIC_DATA_2D

COMPUTED_SPAT_1D_DATA = []
for stat in NP_OPERATORS + OTHER_OPERATORS:
    for name in TEMPORAL_DATA_2D:
        for axis in ["x", "y"]:
            COMPUTED_SPAT_1D_DATA.append(name + "_" + stat + "_" + axis)
TEMPORAL_DATA_1D += COMPUTED_SPAT_1D_DATA

COMPUTED_SPAT_0D_DATA = []
for stat in NP_OPERATORS + OTHER_OPERATORS:
    for name in TEMPORAL_DATA_2D:
        COMPUTED_SPAT_0D_DATA.append(name + "_" + stat + "_xy")
TEMPORAL_DATA_0D += COMPUTED_SPAT_0D_DATA

TEMPORAL_DATA = TEMPORAL_DATA_0D + TEMPORAL_DATA_1D + TEMPORAL_DATA_2D
STATIC_DATA = (
    STATIC_DATA_0D + STATIC_DATA_1D + STATIC_DATA_2D + COMPUTED_STATIC_DATA_2D
)

DATA_NAMES = TEMPORAL_DATA_0D + TEMPORAL_DATA_1D + TEMPORAL_DATA_2D
DATA_NAMES += STATIC_DATA_0D + STATIC_DATA_1D + STATIC_DATA_2D


class AbstractResults:
    """Abstract class for TemporalResults and StaticResults"""

    def __init__(self, name, d, notation=None, **kwargs):
        self.name = name
        self.d = d
        if isinstance(notation, dict):
            self.notation = notations.Notation(**notation)
        elif notation is None:
            self.notation = notations.get_notation(name)
        else:
            self.notation = notation
        self.__dict__.update(kwargs)


class TemporalResults(AbstractResults):
    """Time dependent result of simulation"""

    def __init__(
        self,
        name,
        d,
        t,
        notation=None,
    ):
        super().__init__(name, d, notation=notation)
        # 1d array with times, matching last dimension of self.d
        self.t = t

    def get_temporal_stat(self, stat):
        """Statistical analysis along temporal dimension"""
        if stat in NP_OPERATORS:
            dnew = getattr(np, stat)(self.d, axis=-1)
        elif stat == "final":
            dnew = self.d[..., -1]
        elif stat == "init":
            dnew = self.d[..., 0]
        elif stat == "int":
            dnew = np.trapezoid(self.d, x=self.t)

        notation = notations.add_operator(self.notation, stat, axis="t")

        if dnew.ndim == 2:
            return StaticResults2D(
                self.name + "_" + stat,
                dnew,
                notation=notation,
                x=self.x,
                y=self.y,
                z=self.z,
            )
        elif dnew.ndim == 1:
            return StaticResults1D(
                self.name + "_" + stat,
                dnew,
                notation=notation,
                coords=self.coords,
                coords_name=self.coords_name,
            )

    def get_spatial_stat(self, stat, axis):
        raise NotImplementedError()

    def plot(*arg, **kwargs):
        """Plot results as time dependent"""
        raise NotImplementedError()

    def save(*arg, **kwargs):
        raise NotImplementedError()


class TemporalResults0D(TemporalResults):
    """
    Class inheretied from TemporalResults where the data is one or multiple
    scalar functions of time
    of time.
    """

    def __init__(
        self,
        name,
        d,
        t,
        scalar_names=None,
        notation=None,
    ):
        """
        initiates TemporalResults0D instance

        Parameters
        ----------
        name : string
            Name of the data type.
        d : array
            array like data, with last dimension corresponding to time. It can
            be a one dimensionnal Nt array, or a two dimensionnal NdxNt array,
            where Nt is the legnth of t, and Nd correspond to the number of
            scalar values of interest (e.g. X and Y coordinates of the center
            of mass / front)
        t : array
            Array of time of length Nt.
        scalar_names : list of strings, optional
            List of length Nd containing the names of the scalar fields (one
            name per row of d)

        Returns
        -------
        None.

        """
        super().__init__(name, d, t, notation=notation)
        self.scalar_names = scalar_names

    def plot(self, axe=None, figsize=None, **kwargs):
        """Plot results.

        :param axe: DESCRIPTION, defaults to None
        :type axe: TYPE, optional
        :param figsize: DESCRIPTION, defaults to None
        :type figsize: TYPE, optional
        :param **kwargs: DESCRIPTION
        :type **kwargs: TYPE
        :return: DESCRIPTION
        :rtype: TYPE

        """

        if axe is None:
            fig, axe = plt.subplots(
                1, 1, figsize=figsize, layout="constrained"
            )

        if isinstance(self.d, np.ndarray):
            data = self.d.T
        else:
            data = self.d
        axe.plot(self.t, data, label=self.scalar_names)
        axe.set_xlabel(notations.get_label("t"))
        axe.set_ylabel(notations.get_label(self.notation))

        return axe

    def save(self):
        raise NotImplementedError(
            "Saving method for TemporalResults0D not implemented yet"
        )

    def get_spatial_stat(self, *arg, **kwargs):
        raise NotImplementedError(
            (
                "Spatial integration of Spatialresults0D"
                + " is not implemented because non relevant"
            )
        )


class TemporalResults1D(TemporalResults):
    """Class for simulation results described by one dimension for space
    and one dimension for time.

    :param name: Name of the data
    :type name: str
    :param d: data
    :type d: numpy.ndarray
    :param t: time array
    :type t: array like
    :param coords: 1D coordinate for spatial dimension
    :type coords: array like
    :param coords_name: name of the 1D coordinate (typically "X" or "Y")
    :type coords_name: str

    """

    def __init__(
        self, name, d, t, coords=None, coords_name=None, notation=None
    ):
        """Constructor method."""
        super().__init__(name, d, t, notation=notation)
        # x and y arrays
        self.coords = coords
        self.coords_name = coords_name

    def plot(self, coords=None, **kwargs):
        """Plot results.

        :param axe: DESCRIPTION, defaults to None
        :type axe: TYPE, optional
        :param figsize: DESCRIPTION, defaults to None
        :type figsize: TYPE, optional
        :param **kwargs: DESCRIPTION
        :type **kwargs: TYPE
        :return: DESCRIPTION
        :rtype: TYPE

        """

        if coords is None:
            coords = self.coords

        if coords is None:
            raise TypeError("coords data missing")

        xlabel = notations.get_label(self.coords_name, with_unit=True)
        if "colorbar_kwargs" not in kwargs:
            kwargs["colorbar_kwargs"] = dict()
        if "label" not in kwargs["colorbar_kwargs"]:
            clabel = notations.get_label(self.notation)
            kwargs["colorbar_kwargs"]["label"] = clabel

        axe = plt_fn.plot_shotgather(
            self.coords,
            self.t,
            self.d,
            xlabel=xlabel,
            ylabel=notations.get_label("t"),
            **kwargs
        )

        return axe

    def save(self):
        raise NotImplementedError(
            "Saving method for TemporalResults1D not implemented yet"
        )

    def get_spatial_stat(self, stat, **kwargs):
        """

        :param stat: DESCRIPTION
        :type stat: TYPE
        :return: DESCRIPTION
        :rtype: TYPE

        """

        if stat in NP_OPERATORS:
            dnew = getattr(np, stat)(self.d, axis=0)
        elif stat == "int":
            dd = self.coords[1] - self.coords[0]
            dnew = np.sum(self.d, axis=0) * dd
        notation = notations.add_operator(
            self.notation, stat, axis=self.coords_name
        )
        return TemporalResults0D(
            self.name + "_" + stat, dnew, self.t, notation=notation
        )


class TemporalResults2D(TemporalResults):
    def __init__(self, name, d, t, x=None, y=None, z=None, notation=None):
        """Initiate instance of TemporalResults2D.

        :param name: DESCRIPTION
        :type name: TYPE
        :param d: DESCRIPTION
        :type d: TYPE
        :param t: DESCRIPTION
        :type t: TYPE
        :param x: DESCRIPTION, defaults to None
        :type x: TYPE, optional
        :param y: DESCRIPTION, defaults to None
        :type y: TYPE, optional
        :param z: DESCRIPTION, defaults to None
        :type z: TYPE, optional
        :return: DESCRIPTION
        :rtype: TYPE

        """

        super().__init__(name, d, t, notation=notation)
        # x and y arrays
        self.x = x
        self.y = y
        # topography
        self.z = z

    def plot(
        self,
        x=None,
        y=None,
        z=None,
        file_name=None,
        folder_out=None,
        figsize=None,
        dpi=None,
        fmt="png",
        sup_plt_fn=None,
        sup_plt_fn_args=None,
        **kwargs
    ):
        """Plot results.

        :param **kwargs: DESCRIPTION
        :type **kwargs: TYPE
        :return: DESCRIPTION
        :rtype: TYPE

        """
        if file_name is None:
            file_name = self.name

        if x is None:
            x = self.x
        if y is None:
            y = self.y
        if z is None:
            z = self.z

        if x is None or y is None:
            raise TypeError("x, y or z data missing")

        if z is None:
            warnings.warn("No topography given.")

        if "colorbar_kwargs" not in kwargs:
            kwargs["colorbar_kwargs"] = dict()
        if "label" not in kwargs["colorbar_kwargs"]:
            clabel = notations.get_label(self.notation)
            kwargs["colorbar_kwargs"]["label"] = clabel

        plt_fn.plot_maps(
            x,
            y,
            z,
            self.d,
            self.t,
            file_name=file_name,
            folder_out=folder_out,
            figsize=figsize,
            dpi=dpi,
            fmt=fmt,
            sup_plt_fn=sup_plt_fn,
            sup_plt_fn_args=sup_plt_fn_args,
            **kwargs
        )

        return None

    def save(
        self,
        folder=None,
        file_name=None,
        fmt="asc",
        time=None,
        x=None,
        y=None,
        **kwargs
    ):
        if x is None:
            x = self.x
        if y is None:
            y = self.y
        if x is None or y is None:
            raise ValueError("x et y arrays must not be None")

        if file_name is None:
            file_name = self.name

        if folder is not None:
            file_name = os.path.join(folder, file_name)

        if time is not None:
            if isinstance(time, str):
                if time == "final":
                    inds = [self.d.shape[2] - 1]
                elif time == "initial":
                    inds = [0]
            else:
                inds = [np.argmin(time - np.abs(np.array(self.t) - time))]
        else:
            inds = range(len(self.t))

        for i in inds:
            file_out = file_name + "_{:04d}.".format(i) + fmt
            tilupy.raster.write_raster(
                x, y, self.d[:, :, i], file_out, fmt=fmt, **kwargs
            )

    def get_spatial_stat(self, stat, axis=None):
        """

        :param stat: DESCRIPTION
        :type stat: TYPE
        :param axis: DESCRIPTION, defaults to None
        :type axis: TYPE, optional
        :return: DESCRIPTION
        :rtype: TYPE

        """
        if axis is None:
            axis = (0, 1)

        if isinstance(axis, str):
            axis_str = axis
            if axis == "x":
                axis = 1
            elif axis == "y":
                axis = 0
            elif axis == "xy":
                axis = (0, 1)
        else:
            if axis == 1:
                axis_str = "x"
            elif axis == 0:
                axis_str = "y"
            elif axis == (0, 1):
                axis_str = "xy"

        if stat in NP_OPERATORS:
            dnew = getattr(np, stat)(self.d, axis=axis)
        elif stat == "int":
            dnew = np.sum(self.d, axis=axis)
            if axis == 1:
                dd = self.x[1] - self.x[0]
            elif axis == 0:
                dd = self.y[1] - self.y[0]
            elif axis == (0, 1):
                dd = (self.x[1] - self.x[0]) * (self.y[1] - self.y[0])
            dnew = dnew * dd

        if axis == 1:
            # Needed to get correct orinetation as d[0, 0] is the upper corner
            # of the data, with coordinates x[0], y[-1]
            dnew = np.flip(dnew, axis=0)

        new_name = self.name + "_" + stat + "_" + axis_str
        notation = notations.add_operator(self.notation, stat, axis=axis_str)

        if axis == (0, 1):
            return TemporalResults0D(new_name, dnew, self.t, notation=notation)
        else:
            if axis == 0:
                coords = self.x
                coords_name = "x"
            else:
                coords = self.x
                coords_name = "y"
            return TemporalResults1D(
                new_name,
                dnew,
                self.t,
                coords,
                coords_name=coords_name,
                notation=notation,
            )


class StaticResults(AbstractResults):
    """Result of simulation without time dependence"""

    def __init__(self, name, d, notation=None):
        super().__init__(name, d, notation=notation)
        # x and y arrays

    def plot(self):
        raise NotImplementedError()

    def save(self):
        raise NotImplementedError()


class StaticResults2D(StaticResults):
    def __init__(self, name, d, x=None, y=None, z=None, notation=None):
        super().__init__(name, d, notation=notation)
        # x and y arrays
        self.x = x
        self.y = y
        # topography
        self.z = z

    def plot(
        self,
        axe=None,
        figsize=None,
        x=None,
        y=None,
        z=None,
        sup_plt_fn=None,
        sup_plt_fn_args=None,
        **kwargs
    ):
        """Plot results as map"""

        if axe is None:
            fig, axe = plt.subplots(
                1, 1, figsize=figsize, layout="constrained"
            )

        if x is None:
            x = self.x
        if y is None:
            y = self.y
        if z is None:
            z = self.z

        if x is None or y is None or z is None:
            raise TypeError("x, y or z data missing")

        if "colorbar_kwargs" not in kwargs:
            kwargs["colorbar_kwargs"] = dict()
        if "label" not in kwargs["colorbar_kwargs"]:
            clabel = notations.get_label(self.notation)
            kwargs["colorbar_kwargs"]["label"] = clabel

        axe = plt_fn.plot_data_on_topo(
            x, y, z, self.d, axe=axe, figsize=figsize, **kwargs
        )
        if sup_plt_fn is not None:
            if sup_plt_fn_args is None:
                sup_plt_fn_args = dict()
            sup_plt_fn(axe, **sup_plt_fn_args)

        return axe

    def save(
        self,
        folder=None,
        file_name=None,
        fmt="txt",
        time=None,
        x=None,
        y=None,
        **kwargs
    ):
        if x is None:
            x = self.x
        if y is None:
            y = self.y

        if x is None or y is None:
            raise ValueError("x et y arrays must not be None")

        if file_name is None:
            file_name = self.name + "." + fmt

        if folder is not None:
            file_name = os.path.join(folder, file_name)

        tilupy.raster.write_raster(x, y, self.d, file_name, fmt=fmt, **kwargs)

    def get_spatial_stat(self, stat, axis=None):
        """

        :param stat: DESCRIPTION
        :type stat: TYPE
        :param axis: DESCRIPTION, defaults to None
        :type axis: TYPE, optional
        :return: DESCRIPTION
        :rtype: TYPE

        """
        if axis is None:
            axis = (0, 1)

        if isinstance(axis, str):
            axis_str = axis
            if axis == "x":
                axis = 1
            elif axis == "y":
                axis = 0
            elif axis == "xy":
                axis = (0, 1)
        else:
            if axis == 1:
                axis_str = "x"
            elif axis == 0:
                axis_str = "y"
            elif axis == (0, 1):
                axis_str = "xy"

        if stat in NP_OPERATORS:
            dnew = getattr(np, stat)(self.d, axis=axis)
        elif stat == "int":
            dnew = np.sum(self.d, axis=axis)
            if axis == 1:
                dd = self.x[1] - self.x[0]
            elif axis == 0:
                dd = self.y[1] - self.y[0]
            elif axis == (0, 1):
                dd = (self.x[1] - self.x[0]) * (self.y[1] - self.y[0])
            dnew = dnew * dd

        if axis == 1:
            # Needed to get correct orinetation as d[0, 0] is the upper corner
            # of the data, with coordinates x[0], y[-1]
            dnew = np.flip(dnew, axis=0)

        new_name = self.name + "_" + stat + "_" + axis_str
        notation = notations.add_operator(self.notation, stat, axis=axis_str)

        if axis == (0, 1):
            return StaticResults0D(new_name, dnew, notation=notation)
        else:
            if axis == 0:
                coords = self.x
                coords_name = "x"
            else:
                coords = self.y
                coords_name = "y"
            return StaticResults1D(
                new_name,
                dnew,
                coords,
                coords_name=coords_name,
                notation=notation,
            )


class StaticResults1D(StaticResults):
    def __init__(self, name, d, coords=None, coords_name=None, notation=None):
        """Constructor method."""
        super().__init__(name, d, notation=notation)
        # x and y arrays
        self.coords = coords
        self.coords_name = coords_name

    def plot(self, axe=None, figsize=None, **kwargs):
        """Plot results.

        :param axe: DESCRIPTION, defaults to None
        :type axe: TYPE, optional
        :param figsize: DESCRIPTION, defaults to None
        :type figsize: TYPE, optional
        :param **kwargs: DESCRIPTION
        :type **kwargs: TYPE
        :return: DESCRIPTION
        :rtype: TYPE

        """

        if axe is None:
            fig, axe = plt.subplots(
                1, 1, figsize=figsize, layout="constrained"
            )

        if isinstance(self.d, np.ndarray):
            data = self.d.T
        else:
            data = self.d
        axe.plot(self.coords, data, label=self.scalar_names)
        axe.set_xlabel(notations.get_label(self.coords_name))
        axe.set_ylabel(notations.get_label(self.notation))

        return axe


class StaticResults0D(StaticResults):
    def __init__(self, d, name, notation=None):
        super().__init__(name, d, notation=notation)


class Results:
    """Results of thin-layer model simulation

    This class is the parent class for all simulation results, whatever the
    kind of input data. Methods and functions for processing results are given
    here. Reading results from code specific outputs is done in inhereited
    classes.
    """

    def __init__(self, *args, **kwargs):
        """
        Call init function used for code in inherited classes.

        It is not recommended

        Parameters
        ----------
        code : string
            Name of code from which results must be read.
        """
        self._h_max = None
        self._h = None
        self._costh = None
        self._zinit = None
        self.folder_output = None
        self.tim = None

    @property
    def zinit(self):
        """Get initial topography"""
        return self._zinit

    @property
    def z(self):
        """Alias for zinit"""
        return self.zinit

    @property
    def h(self):
        """Get thickness"""
        if self._h is None:
            self._h = self.get_output("h").d
        return self._h

    @property
    def h_max(self):
        """Get maximum thickness"""
        if self._h_max is None:
            self._h_max = self.get_output("h_max").d
        return self._h_max

    def get_costh(self):
        """Get cos(slope) of topography"""
        [Fx, Fy] = np.gradient(self.zinit, np.flip(self.y), self.x)
        costh = 1 / np.sqrt(1 + Fx**2 + Fy**2)
        return costh

    @property
    def costh(self):
        """Compute or get cos(slope) of topography"""
        if self._costh is None:
            self._costh = self.get_costh()
        return self._costh

    def get_output(self, output_name, from_file=True, **kwargs):
        # Specific case of center of mass
        if output_name == "centermass":
            return self.get_center_of_mass(**kwargs)

        strs = output_name.split("_")
        n_strs = len(strs)
        # If no operator is called, call directly _get_output
        if n_strs == 1:
            res = self._get_output(output_name, **kwargs)
            return res

        # Otherwise, get name, operator and axis (optional)
        name = strs[0]
        operator = strs[1]
        if n_strs == 3:
            axis = strs[2]
        else:
            axis = None

        res = None
        # If processed output is read directly from file, call the child method
        # _read_from_file.
        if from_file:
            res = self._read_from_file(name, operator, axis=axis, **kwargs)
            # res is None in case of function failure

        # If no results could be read from file, output must be
        # processed by tilupy
        if res is None:
            # Get output from name
            res = self._get_output(name, x=self.x, y=self.y, **kwargs)
            if axis is None:
                # If no axis is given, the operator operates over time by
                # default
                res = res.get_temporal_stat(operator)
            else:
                if axis == "t":
                    res = res.get_temporal_stat(operator)
                else:
                    res = res.get_spatial_stat(operator, axis=axis)

        return res

    def _get_output(self):
        raise NotImplementedError

    def _read_from_file(self):
        raise NotImplementedError

    def get_center_of_mass(self, h_thresh=None):
        """
        Compute center of mass coordinates depending on time
        :param h_thresh: DESCRIPTION, defaults to None
        :type h_thresh: TYPE, optional
        :return: DESCRIPTION
        :rtype: TYPE

        """
        dx = self.x[1] - self.x[0]
        dy = self.y[1] - self.y[0]
        # Make meshgrid
        X, Y = np.meshgrid(self.x, np.flip(self.y))

        # Weights for coordinates average (volume in cell / total volume)
        h2 = self.h.copy()
        if h_thresh is not None:
            h2[h2 < h_thresh] = 0
        w = h2 / self.costh[:, :, np.newaxis] * dx * dy
        vol = np.nansum(w, axis=(0, 1))
        w = w / vol[np.newaxis, np.newaxis, :]
        # Compute center of mass coordinates
        nt = h2.shape[2]
        coord = np.zeros((3, nt))
        tmp = X[:, :, np.newaxis] * w
        coord[0, :] = np.nansum(tmp, axis=(0, 1))
        tmp = Y[:, :, np.newaxis] * w
        coord[1, :] = np.nansum(tmp, axis=(0, 1))
        tmp = self.zinit[:, :, np.newaxis] * w
        coord[2, :] = np.nansum(tmp, axis=(0, 1))
        # Make TemporalResults
        res = TemporalResults0D(
            "centermass",
            coord,
            self.tim,
            scalar_names=["X", "Y", "z"],
            notation=None,
        )
        return res

    def plot(
        self,
        name,
        save=True,
        folder_out=None,
        dpi=150,
        fmt="png",
        file_suffix=None,
        file_prefix=None,
        h_thresh=None,
        from_file=True,
        display_plot=True,
        **kwargs
    ):
        """


        Parameters
        ----------
        name : TYPE
            DESCRIPTION.
        save : TYPE, optional
            DESCRIPTION. The default is True.
        folder_out : TYPE, optional
            DESCRIPTION. The default is None.
        dpi : TYPE, optional
            DESCRIPTION. The default is 150.
        fmt : TYPE, optional
            DESCRIPTION. The default is "png".
        h_thresh : TYPE, optional
            DESCRIPTION. The default is None.
        from_file : TYPE, optional
            DESCRIPTION. The default is False.
        display_plot : TYPE, optional
            DESCRIPTION. The default is True.
        **kwargs : TYPE
            DESCRIPTION.

        Returns
        -------
        axe : TYPE
            DESCRIPTION.
        fig : TYPE
            DESCRIPTION.

        """

        if not display_plot:
            backend = plt.get_backend()
            plt.close("all")
            plt.switch_backend("Agg")

        data = self.get_output(name, from_file=from_file, h_thresh=h_thresh)

        # if name in TEMPORAL_DATA_2D + STATIC_DATA_2D:
        #     if "colorbar_kwargs" not in kwargs:
        #         kwargs["colorbar_kwargs"] = dict()
        #     if "label" not in kwargs["colorbar_kwargs"]:
        #         kwargs["colorbar_kwargs"]["label"] = notations.get_label(
        #             self.notation
        #         )

        # if "x" not in kwargs:
        #     kwargs["x"] = self.x
        # if "y" not in kwargs:
        #     kwargs["y"] = self.y
        # if "z" not in kwargs:
        #     kwargs["z"] = self.zinit

        if save:
            if folder_out is None:
                assert (
                    self.folder_output is not None
                ), "folder_output attribute must be set"
                folder_out = os.path.join(self.folder_output, "plots")
            os.makedirs(folder_out, exist_ok=True)

        if folder_out is not None and isinstance(data, TemporalResults2D):
            # If data is TemporalResults2D then saving is managed directly
            # by the associated plot method
            kwargs["folder_out"] = folder_out
            kwargs["dpi"] = dpi
            kwargs["fmt"] = fmt
            # kwargs["file_suffix"] = file_prefix
            # kwargs["file_prefix"] = file_prefix

        axe = data.plot(**kwargs)

        if folder_out is not None and not isinstance(data, TemporalResults2D):
            file_name = name
            if file_suffix is not None:
                file_name = file_name + "_" + file_suffix
            if file_prefix is not None:
                file_name = file_prefix + "_" + file_name
            file_out = os.path.join(folder_out, file_name + "." + fmt)
            # axe.figure.tight_layout(pad=0.1)
            axe.figure.savefig(
                file_out, dpi=dpi, bbox_inches="tight", pad_inches=0.05
            )

        if not display_plot:
            plt.close("all")
            plt.switch_backend(backend)

        return axe

    def save(
        self,
        name,
        folder=None,
        file_name=None,
        fmt="txt",
        from_file=True,
        **kwargs
    ):
        if folder is None:
            assert (
                self.folder_output is not None
            ), "folder_output attribute must be set"
            folder = os.path.join(self.folder_output, "processed")
            os.makedirs(folder, exist_ok=True)

        if name in DATA_NAMES:
            data = self.get_output(name, from_file=from_file)
            if data.d.ndim > 1:
                if "x" not in kwargs:
                    kwargs["x"] = self.x
                if "y" not in kwargs:
                    kwargs["y"] = self.y

            data.save(folder=folder, file_name=file_name, fmt=fmt, **kwargs)

        elif name in TOPO_DATA_2D:
            if file_name is None:
                file_name = name
            file_out = os.path.join(folder, file_name)
            tilupy.raster.write_raster(
                self.x,
                self.y,
                getattr(self, name),
                file_out,
                fmt=fmt,
                **kwargs
            )


def get_results(code, **kwargs):
    """
    Get simulation results for a given code. This function calls the
    appropriate module.

    Parameters
    ----------
    code : TYPE
        DESCRIPTION.
    **kwargs : TYPE
        DESCRIPTION.

    Returns
    -------
    None.

    """
    module = importlib.import_module("tilupy.models." + code + ".read")
    return module.Results(**kwargs)


def use_thickness_threshold(simu, array, h_thresh):
    thickness = simu.get_output("h")
    array[thickness.d < h_thresh] = 0
    return array



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 21 12:10:46 2021

@author: peruzzetto
"""
import os
import numpy as np

import tilupy.make_topo
import tilupy.make_mass
import tilupy.plot


def make_constant_slope(
    folder_out,
    theta=10,
    m_radius=50,
    m_height=50,
    m_x=None,
    m_y=None,
    xmax=800,
    ymax=550,
    cellsize=2.5,
):
    """Init topography and mass for simulation on constant slope."""
    # Create mesh
    x = np.linspace(0, xmax, int(np.round(xmax / cellsize)))
    y = np.linspace(0, ymax, int(np.round(ymax / cellsize)))
    xmesh, ymesh = np.meshgrid(x, y)

    # Topography slope
    slope = np.tan(np.deg2rad(theta))

    # Topography array
    z = -slope * (xmesh - xmax)

    # Create initial mass
    if m_x is None:
        m_x = 3 * m_radius
    if m_y is None:
        m_y = ymax / 2
    m = (
        1
        - (xmesh - m_x) ** 2 / m_radius**2
        - (ymesh - m_y) ** 2 / m_radius**2
    )
    m = np.maximum(m * m_height, 0)

    # Write headers for ascii files
    nx = z.shape[1]
    ny = z.shape[0]
    header_txt = "ncols {:.0f}\nnrows {:.0f}\nxllcorner 0\nyllcorner 0\n"
    header_txt += "cellsize {:.4f}\nnodata_value -99999"
    header_txt = header_txt.format(nx, ny, cellsize)

    for a, name in zip([z, m], ["topo", "mass"]):
        np.savetxt(
            os.path.join(folder_out, name + ".asc"),
            a,
            header=header_txt,
            comments="",
        )


def gray99_topo_mass(
    dx=0.1, dy=0.1, save=False, folder_out=None, res_type="true_normal"
):
    # Initiate topography
    X, Y, Z = tilupy.make_topo.gray99(dx=dx, dy=dy)

    # Initiate initial mass. It is a spherical calotte above the topography,
    # in Gray et al 99 (p. 1859) the resulting mass has a height of 0.22 m and a radius
    # of 0.32 m (more precisely it is the length in the downslope direction)
    # The correspondig radius of the sphere, and the offset from the topography
    # in the topography normal direction (norm_offset) are deduced from these
    # parameters. See also Gig 3 in Wieland, Gray and Hutter (1999)

    x0 = 0.06 * np.cos(np.deg2rad(40))
    hmass = 0.22
    wmass = 0.32
    radius = (wmass**2 + hmass**2) / (2 * hmass)
    norm_offset = (wmass**2 - hmass**2) / (2 * hmass)
    # Z = -np.tile(X, [len(Y), 1])*np.tan(np.deg2rad(20))
    M = tilupy.make_mass.calotte(
        X, Y, Z, x0, 0, radius, norm_offset=norm_offset, res_type=res_type
    )

    return X, Y, Z, M


if __name__ == "__main__":
    x, y, z, m = gray99_topo_mass(dx=0.01, dy=0.01)
    tilupy.plot.plot_data_on_topo(x, y, z, m, topo_kwargs=dict(level_min=0.1))



# -*- coding: utf-8 -*-
"""
Created on Mon Oct 10 14:47:17 2022

@author: peruzzetto
"""

import numpy as np
import matplotlib.pyplot as plt

import shapely.geometry as geom
import shapely.ops


def CSI(pred, obs):
    ipred = pred > 0
    iobs = obs > 0

    TP = np.sum(ipred * iobs)
    FP = np.sum(ipred * ~iobs)
    FN = np.sum(~ipred * iobs)

    return TP / (TP + FP + FN)


def diff_runout(
    x_contour, y_contour, point_ref, section=None, orientation="W-E"
):
    npts = len(x_contour)
    contour = geom.LineString(
        [(x_contour[i], y_contour[i]) for i in range(npts)]
    )
    point = geom.Point(point_ref)
    if section is None:
        return point.distance(contour)
    elif isinstance(section, np.ndarray):
        section = geom.LineString(section)

    assert isinstance(section, geom.LineString)
    section = revert_line(section, orientation)
    intersections = section.intersection(contour)
    if isinstance(intersections, geom.MultiPoint):
        intersections = geom.LineString(section.intersection(contour))
    intersections = np.array(intersections.coords)
    if orientation == "W-E":
        i = np.argmax(intersections[:, 0])
    if orientation == "E-W":
        i = np.argmin(intersections[:, 0])
    if orientation == "S-N":
        i = np.argmax(intersections[:, 1])
    if orientation == "N-S":
        i = np.argmin(intersections[:, 1])
    intersection = geom.Point(intersections[i, :])

    #######
    # plt.figure()
    # cont = np.array(contour.coords)
    # sec = np.array(section.coords)
    # inter = np.array(intersection.coords)
    # pt = np.array(point.coords)
    # plt.plot(cont[:,0], cont[:,1],
    #           sec[:,0], sec[:,1],
    #           pt[:,0], pt[:,1], 'o',
    #           inter[:,0], inter[:,1],'x')
    #######

    return section.project(intersection) - section.project(point)


def revert_line(line, orientation="W-E"):
    pt_init = line.coords[0]
    pt_end = line.coords[-1]
    if orientation == "W-E":
        if pt_init[0] > pt_end[0]:
            line = shapely.ops.substring(line, 1, 0, normalized=True)
    elif orientation == "E-W":
        if pt_init[0] < pt_end[0]:
            line = shapely.ops.substring(line, 1, 0, normalized=True)
    if orientation == "S-N":
        if pt_init[1] > pt_end[1]:
            line = shapely.ops.substring(line, 1, 0, normalized=True)
    elif orientation == "N-S":
        if pt_init[1] < pt_end[1]:
            line = shapely.ops.substring(line, 1, 0, normalized=True)

    return line


def get_contour(x, y, z, zlevels, indstep=1, maxdist=30, closed_contour=True):
    # Add sup_value at the border of the array, to make sure contour
    # lines are closed
    if closed_contour:
        z2 = z.copy()
        ni = z2.shape[0]
        nj = z2.shape[1]
        z2 = np.vstack([np.zeros((1, nj)), z2, np.zeros((1, nj))])
        z2 = np.hstack([np.zeros((ni + 2, 1)), z2, np.zeros((ni + 2, 1))])
        dxi = x[1] - x[0]
        dxf = x[-1] - x[-2]
        dyi = y[1] - y[0]
        dyf = y[-1] - y[-2]
        x2 = np.insert(np.append(x, x[-1] + dxf), 0, x[0] - dxi)
        y2 = np.insert(np.append(y, y[-1] + dyf), 0, y[0] - dyi)
    else:
        x2, y2, z2 = x, y, z

    backend = plt.get_backend()
    plt.switch_backend("Agg")
    plt.figure()
    ax = plt.gca()
    cs = ax.contour(x2, y2, np.flip(z2, 0), zlevels)
    nn1 = 1
    v1 = np.zeros((1, 2))
    xcontour = {}
    ycontour = {}
    for indlevel in range(len(zlevels)):
        for p in cs.collections[indlevel].get_paths():
            if p.vertices.shape[0] > nn1:
                v1 = p.vertices
                nn1 = p.vertices.shape[0]
        xc = [v1[::indstep, 0]]
        yc = [v1[::indstep, 1]]
        if maxdist is not None and not closed_contour:
            ddx = np.abs(v1[0, 0] - v1[-1, 0])
            ddy = np.abs(v1[0, 1] - v1[-1, 1])
            dd = np.sqrt(ddx**2 + ddy**2)
            if dd > maxdist:
                xc[0] = None
                yc[0] = None
        xcontour[zlevels[indlevel]] = xc[0]
        ycontour[zlevels[indlevel]] = yc[0]
    plt.switch_backend(backend)
    return xcontour, ycontour


def format_path_linux(path):
    """
    Change a Windows-type path to a path formatted for Linux. \\ are changed
    to /, and partitions like "C:" are changed to "/mnt/c/"

    Parameters
    ----------
    path : string
        String with the path to be modified.

    Returns
    -------
    path2 : string
        Formatted path.

    """
    if path[1] == ":":
        path2 = "/mnt/{:s}/".format(path[0].lower()) + path[2:]
    else:
        path2 = path
    path2 = path2.replace("\\", "/")
    if " " in path2:
        path2 = '"' + path2 + '"'
    return path2



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun  1 15:27:36 2021

@author: peruzzetto
"""
import os

import numpy as np
import tilupy.read

from tilupy import notations

# Dictionnary with results names lookup table, to match code output names
LOOKUP_NAMES = dict(
    h="rho",
    hvert="rho",
    ux="u",
    uy="ut",
    u="unorm",
    hu="momentum",
    ek="ek",
    vol="vol",
    ep="ep",
    etot="etot",
)

# Classify results
STATES_OUTPUT = ["h", "ux", "uy", "hvert"]

tmp = ["facc", "fcurv", "ffric", "fgrav", "finert", "fpression"]

FORCES_OUTPUT = []
for force in tmp:
    FORCES_OUTPUT.append(force + "x")
    FORCES_OUTPUT.append(force + "y")

FORCES_OUTPUT += [
    "shearx",
    "sheary",
    "shearz",
    "normalx",
    "normaly",
    "normalz",
    "pbottom",
    "pcorrdt",
    "pcorrdiv",
]

INTEGRATED_OUTPUT = ["ek", "ep", "etot"]


def read_params(file):
    """
    Read simulation parameters from file.

    Parameters
    ----------
    file : str
        Parameters file.

    Returns
    -------
    params : dict
        Dictionnary with parameters

    """
    if file is None:
        params = None
    else:
        params = dict()
        with open(file, "r") as f:
            lines = filter(None, (line.rstrip() for line in f))
            for line in lines:
                (key, val) = line.split(" ")
                try:
                    params[key] = float(val)
                except ValueError:
                    params[key] = val.rstrip()
        params["nx"] = int(params["nx"])
        params["ny"] = int(params["ny"])

    return params


def get_axes(**varargs):
    """Set x and y axes."""
    if "x0" in varargs:
        x0 = varargs["x0"]
    else:
        x0 = 0
    if "y0" in varargs:
        y0 = varargs["y0"]
    else:
        y0 = 0

    nx = varargs["nx"]
    ny = varargs["ny"]
    dx = varargs["per"] / nx
    dy = varargs["pery"] / ny
    x = dx * np.arange(nx) + dx / 2
    y = dy * np.arange(ny) + dy / 2

    try:
        coord_pos = varargs["coord_pos"]
    except KeyError:
        coord_pos = "bottom_left"

    if coord_pos == "bottom_left":
        x = x + x0
        y = y + y0
    elif coord_pos == "upper_right":
        x = x + x0 - x[-1]
        y = y + y0 - y[-1]
    elif coord_pos == "upper_left":
        x = x + x0
        y = y + y0 - y[-1]
    elif coord_pos == "lower_right":
        x = x + x0 - x[-1]
        y = y + y0

    return x, y


def read_file_bin(file, nx, ny):
    """Read shaltop .bin result file."""
    data = np.fromfile(file, dtype=np.float32)
    nbim = int(np.size(data) / nx / ny)
    data = np.reshape(data, (nx, ny, nbim), order="F")
    data = np.transpose(np.flip(data, axis=1), (1, 0, 2))

    return data


def read_file_init(file, nx, ny):
    """Read shaltop initial .d files."""
    data = np.loadtxt(file)
    data = np.reshape(data, (nx, ny), order="F")
    data = np.transpose(np.flip(data, axis=1), (1, 0))
    return data


class Results(tilupy.read.Results):
    """Results of shaltop simulations."""

    def __init__(self, file_params=None, folder_base=None, **varargs):
        """
        Init simulation results.

        Parameters
        ----------
        file_params : str
            File where simulation parameters will be read

        """
        super().__init__()

        if folder_base is None:
            folder_base = os.getcwd()
        if file_params is None:
            file_params = "params.txt"

        if "." not in file_params:
            file_params = file_params + ".txt"

        file_params = os.path.join(folder_base, file_params)

        params = read_params(file_params)
        x, y = get_axes(**params)

        varargs.update(
            dict(
                code="shaltop",
                htype="normal",
                params=params,
                x=x,
                y=y,
                nx=len(x),
                ny=len(y),
            )
        )

        for key in varargs:
            setattr(self, key, varargs[key])

        self.folder_base = folder_base
        # Folder where results are stored
        if "folder_output" not in self.params:
            self.folder_output = os.path.join(self.folder_base, "data2")
        else:
            self.folder_output = os.path.join(
                self.folder_base, self.params["folder_output"]
            )

        # Get time of outputs
        self.tim = np.loadtxt(os.path.join(self.folder_output, "time_im.d"))
        file_tforces = os.path.join(self.folder_output, "time_forces.d")
        if os.path.isfile(file_tforces):
            self.tforces = np.loadtxt(
                os.path.join(self.folder_output, "time_forces.d")
            )
        else:
            self.tforces = []

    @property
    def zinit(self):
        """Compute or get cos(slope) of topography"""
        if self._zinit is None:
            self.set_zinit()
        return self._zinit

    def set_zinit(self, zinit=None):
        """Set zinit, initial topography."""
        path_zinit = os.path.join(self.folder_output, "z.bin")
        if not os.path.isfile(path_zinit) and "file_z_init" in self.params:
            path_zinit = os.path.join(
                self.folder_base, self.params["file_z_init"]
            )
            self._zinit = read_file_init(path_zinit, self.nx, self.ny)
        else:
            self._zinit = np.squeeze(
                read_file_bin(path_zinit, self.nx, self.ny)
            )

    def _read_from_file(self, name, operator, axis=None, **kwargs):
        res = None

        if name in ["u", "momentum", "h"]:
            if operator in ["max"] and axis in [None, "t"]:
                file = os.path.join(
                    self.folder_output, LOOKUP_NAMES[name] + operator + ".bin"
                )
                d = np.squeeze(read_file_bin(file, self.nx, self.ny))
                res = tilupy.read.StaticResults2D(
                    "_".join([name, operator]), d, x=self.x, y=self.y, z=self.z
                )

        if (name, operator) == ("hu2", "int"):
            array = np.loadtxt(os.path.join(self.folder_output, "ek.d"))
            d = array[:, 1]
            t = array[:, 0]
            res = tilupy.read.TemporalResults0D(name, d, t)

        return res

    def _get_output(self, name, **kwargs):
        # Read thicknesses or velocity components
        d = None
        t = None
        notation = None

        if name in STATES_OUTPUT:
            file = os.path.join(
                self.folder_output, LOOKUP_NAMES[name] + ".bin"
            )
            d = read_file_bin(file, self.nx, self.ny)
            if name == "hvert":
                d = d / self.costh[:, :, np.newaxis]
            if name in ["ux", "uy"]:
                d = d * self.get_costh()[:, :, np.newaxis]
            t = self.tim

        if name == "u":
            d = self.get_u()
            t = self.tim

        if name in ["hu", "hu2"]:
            fileh = os.path.join(self.folder_output, "rho.bin")
            h = read_file_bin(fileh, self.nx, self.ny)
            u = self.get_u()
            if name == "hu":
                d = h * u
            elif name == "hu2":
                d = h * u**2
            t = self.tim

        if name in INTEGRATED_OUTPUT:
            array = np.loadtxt(
                os.path.join(self.folder_output, LOOKUP_NAMES[name] + ".d")
            )
            d = array[:, 1]
            if "density" in self.params:
                density = self.params["density"]
            else:
                density = 1
            d = d * density
            t = array[:, 0]

        if name in FORCES_OUTPUT:
            file = os.path.join(self.folder_output, name + ".bin")
            d = read_file_bin(file, self.nx, self.ny)
            t = self.tforces
            notation = notations.Notation(
                name,
                long_name=name,
                unit=notations.Unit(Pa=1, kg=-1, m=3),
                symbol=name,
            )

        if d is None:
            file = os.path.join(self.folder_output, name)
            if os.path.isfile(file + ".bin"):
                d = read_file_bin(file + ".bin", self.nx, self.ny)
                t = self.tim
            elif os.path.isfile(file + ".d"):
                d = np.loadtxt(file + ".d")

        if (
            "h_thresh" in kwargs
            and kwargs["h_thresh"] is not None
            and d.ndim == 3
        ):
            d = tilupy.read.use_thickness_threshold(
                self, d, kwargs["h_thresh"]
            )

        if t is None:
            return tilupy.read.AbstractResults(name, d, notation=notation)

        else:
            if d.ndim == 3:
                return tilupy.read.TemporalResults2D(
                    name, d, t, notation=notation, x=self.x, y=self.y, z=self.z
                )
            if d.ndim == 1:
                return tilupy.read.TemporalResults0D(
                    name, d, t, notation=notation
                )

        return None

    # def get_temporal_output(self, name, d=None, t=None, **varargs):
    #     """
    #     Read 2D time dependent simulation results.

    #     Parameters
    #     ----------
    #     name : str
    #         Name of output.
    #     d : ndarray, optional
    #         Data to be read. If None, will be computed.
    #         The default is None.
    #     t : ndarray, optional
    #         Time of results snapshots (1D array), matching last dimension of d.
    #         If None, will be computed. The default is None.
    #     **varargs : TYPE
    #         DESCRIPTION.

    #     """
    #     # Read thicknesses or velocity components
    #     if name in STATES_OUTPUT:
    #         file = os.path.join(
    #             self.folder_output, LOOKUP_NAMES[name] + ".bin"
    #         )
    #         d = read_file_bin(file, self.nx, self.ny)
    #         if name == "hvert":
    #             d = d / self.costh[:, :, np.newaxis]
    #         t = self.tim

    #     # Raed integrated kinetic energy
    #     if name in ["hu2_int"]:
    #         array = np.loadtxt(
    #             os.path.join(self.folder_output, LOOKUP_NAMES[name] + ".d")
    #         )
    #         d = array[:, 1]
    #         t = array[:, 0]

    #     # Compute the velocity
    #     if name == "u":
    #         d = self.get_u()
    #         t = self.tim

    #     if name in ["hu", "hu2"]:
    #         fileh = os.path.join(self.folder_output, "rho.bin")
    #         h = read_file_bin(fileh, self.nx, self.ny)
    #         u = self.get_u()
    #         if name == "hu":
    #             d = h * u
    #         elif name == "hu2":
    #             d = h * u**2
    #         t = self.tim

    #     if (
    #         "h_thresh" in varargs
    #         and varargs["h_thresh"] is not None
    #         and d.ndim == 3
    #     ):
    #         d = tilupy.read.use_thickness_threshold(
    #             self, d, varargs["h_thresh"]
    #         )

    #     return tilupy.read.TemporalResults2D(name, d, t)

    # def get_static_output(self, name, d=None, from_file=True, **varargs):
    #     """
    #     Read 2D time dependent simulation results.

    #     Parameters
    #     ----------
    #     name : str
    #         Name of output.
    #     d : ndarray, optional
    #         Data to be read. Last dimension is for time.
    #         If None, will be computed.
    #         The default is None.
    #     **varargs : TYPE
    #         DESCRIPTION.

    #     """
    #     if name in tilupy.read.COMPUTED_STATIC_DATA_2D:
    #         state, stat = name.split("_")
    #         if stat in ["final", "initial"]:
    #             hh = self.get_temporal_output(state)
    #             if state == "hvert":
    #                 hh.d = hh.d / self.costh[:, :, np.newaxis]
    #             return hh.get_temporal_stat(stat)
    #         if from_file:
    #             file = os.path.join(
    #                 self.folder_output, LOOKUP_NAMES[state] + stat + ".bin"
    #             )
    #             if os.path.isfile(file):
    #                 d = np.squeeze(read_file_bin(file, self.nx, self.ny))
    #             else:
    #                 print(
    #                     file
    #                     + " was not found, "
    #                     + name
    #                     + "_"
    #                     + stat
    #                     + " computed from temporal output."
    #                 )
    #                 from_file = False
    #         if not from_file:
    #             data = self.get_temporal_output(state)
    #             if state == "hvert":
    #                 hh.d = hh.d / self.costh
    #             d = data.get_temporal_stat(stat).d
    #     else:
    #         raise (NotImplementedError())

    #     return tilupy.read.StaticResults(name, d)

    def get_u(self):
        """Compute velocity norm from results"""
        file = os.path.join(self.folder_output, "u" + ".bin")
        u = read_file_bin(file, self.nx, self.ny)
        file = os.path.join(self.folder_output, "ut" + ".bin")
        ut = read_file_bin(file, self.nx, self.ny)

        [Fx, Fy] = np.gradient(self.zinit, np.flip(self.y), self.x)
        u = u * self.costh[:, :, np.newaxis]
        ut = ut * self.costh[:, :, np.newaxis]
        d = np.sqrt(
            u**2
            + ut**2
            + (Fx[:, :, np.newaxis] * u + Fy[:, :, np.newaxis] * ut) ** 2
        )

        return d



# -*- coding: utf-8 -*-
"""
Created on Wed Jun  5 11:07:44 2024

@author: peruzzetto
"""

import os

import numpy as np
import tilupy.read

from scipy.interpolate import RegularGridInterpolator


class Results(tilupy.read.Results):
    def __init__(self, folder, name, raster, grid="edges"):
        super().__init__()

        self.folder = folder
        self.folder_output = folder  # Results are in the same folder
        # than simulation data
        self.name = name
        # grid=='edges' -> use edges, output will have same dimension as
        # topography raster
        # grid=='cell' -> use cell, output nx and ny sizes are the size of
        # topography raster, minus 1
        self.grid = grid

        if not raster.endswith(".asc"):
            raster = raster + ".asc"
        self.raster = raster

        self.params = dict()

        self.x, self.y, self._zinit = tilupy.raster.read_ascii(
            os.path.join(folder, self.raster)
        )
        self.dx = self.x[1] - self.x[0]
        self.dy = self.y[1] - self.y[0]
        if self.grid == "cells":
            fz = RegularGridInterpolator(
                (self.y, self.x),
                self._zinit,
                method="linear",
                bounds_error=False,
                fill_value=None,
            )
            self.x = self.x[1:] - self.dx / 2
            self.y = self.y[1:] - self.dy / 2
            x_mesh, y_mesh = np.meshgrid(self.x, self.y)
            self._zinit = fz((y_mesh, x_mesh))

        self.params["nx"] = len(self.x)
        self.params["ny"] = len(self.y)

        # Rheology
        rheol = np.loadtxt(os.path.join(folder, self.name + ".rhe"))
        self.params["tau/rho"] = rheol[0]
        self.params["K/tau"] = rheol[1]

        # Numerical parameters
        with open(os.path.join(self.folder, "DONLEDD1.DON"), "r") as fid:
            for line in fid:
                name = line[:34].strip(" ")
                value = line[34:]
                if len(value) == 1:
                    value = int(value)
                else:
                    value = float(value)
                self.params[name] = value

    def read_resfile(
        self,
    ):
        # read initial mass
        h_init = np.loadtxt(
            os.path.join(self.folder, self.name + ".cin"), skiprows=1
        )

        # Read results
        file_res = os.path.join(self.folder, self.name + ".asc")
        n_times = int(self.params["tmax"] / self.params["dtsorties"])
        self._tim = [0]
        if self.grid == "edges":
            ny_out = self.params["ny"] - 1
            x_out = self.x[1:] - self.dx / 2
            y_out = self.y[1:] - self.dy / 2
            x_mesh, y_mesh = np.meshgrid(self.x, self.y)
            h_init = h_init.reshape(
                self.params["ny"] - 1, self.params["nx"] - 1
            )
            fh = RegularGridInterpolator(
                (y_out, x_out),
                h_init,
                method="linear",
                bounds_error=False,
                fill_value=None,
            )
            h_init = fh((y_mesh, x_mesh))
        else:
            ny_out = self.params["ny"]
            h_init = h_init.reshape(self.params["ny"], self.params["nx"])
        # x_mesh_out, y_mesh_out = np.meshgrid(
        #     self.x[1:] - self.dx / 2, self.y[1:] - self.dy / 2
        # )

        self._h = np.zeros((self.params["ny"], self.params["nx"], n_times + 1))
        self._h[:, :, 0] = np.flip(h_init, axis=0)
        self._u = np.zeros((self.params["ny"], self.params["nx"], n_times + 1))
        with open(file_res, "r") as fid:
            lines = fid.readlines()
            n_lines = 2 * (ny_out) + 8  # Number of lines per time step
            for i_time in range(n_times):
                i_start = n_lines * i_time
                i_start_h = i_start + 7
                i_stop_h = i_start + ny_out + 7
                i_start_u = i_start + ny_out + 8
                i_stop_u = i_start + 2 * ny_out + 8
                self._tim.append(float(lines[i_start]))
                h_out = np.loadtxt(lines[i_start_h:i_stop_h])
                u_out = np.loadtxt(lines[i_start_u:i_stop_u])
                if self.grid == "edges":
                    fh = RegularGridInterpolator(
                        (y_out, x_out),
                        h_out,
                        method="linear",
                        bounds_error=False,
                        fill_value=None,
                    )
                    self._h[:, :, i_time + 1] = fh((y_mesh, x_mesh))
                    fu = RegularGridInterpolator(
                        (y_out, x_out),
                        u_out,
                        method="linear",
                        bounds_error=False,
                        fill_value=None,
                    )
                    self._u[:, :, i_time + 1] = fu((y_mesh, x_mesh))
                else:
                    self._h[:, :, i_time + 1] = h_out
                    self._u[:, :, i_time + 1] = u_out

    @property
    def h(self):
        if self._h is None:
            self.read_resfile()
        return self._h

    @property
    def u(self):
        if self._u is None:
            self.read_resfile()
        return self._u

    @property
    def tim(self):
        if self._tim is None:
            self.read_resfile()
        return self._tim

    def _get_output(self, name, **kwargs):
        d = None
        t = None
        notation = None

        if name in ["h", "u"]:
            d = getattr(self, name)
            t = self.tim
            return tilupy.read.TemporalResults2D(
                name, d, t, notation=notation, x=self.x, y=self.y, z=self.z
            )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May 25 15:18:31 2021

@author: peruzzetto
"""

import os
import posixpath
import numpy as np

from tilupy.utils import format_path_linux

import tilupy.notations
import tilupy.raster


README_PARAM_MATCH = dict(
    tmax="tmax", CFL="cflhyp", h_min="eps0", dt_im_output="dt_im"
)

SHALTOP_LAW_ID = dict(coulomb=1, voellmy=8, bingham=6, muI=7)


def write_params_file(params, directory=None, file_name="params.txt"):
    """
    Write params file for shaltop simulations

    Parameters
    ----------
    params : TYPE
        DESCRIPTION.
    sup_data : TYPE, optional
        DESCRIPTION. The default is {}.
    directory : TYPE, optional
        DESCRIPTION. The default is None.
    file_name : TYPE, optional
        DESCRIPTION. The default is 'params.txt'.

    Returns
    -------
    None.

    """

    if directory is None:
        directory = os.getcwd()
    with open(os.path.join(directory, file_name), "w") as file_params:
        for name in params:
            val = params[name]
            if isinstance(val, int) or isinstance(val, np.int64):
                file_params.write("{:s} {:d}\n".format(name, val))
            if isinstance(val, float) or isinstance(val, np.float64):
                file_params.write("{:s} {:.8G}\n".format(name, val))
            if isinstance(val, str):
                file_params.write("{:s} {:s}\n".format(name, val))


def raster_to_shaltop_txtfile(file_in, file_out, folder_out=None):
    if folder_out is not None:
        file_out = os.path.join(folder_out, file_out)

    x, y, rast = tilupy.raster.read_raster(file_in)
    np.savetxt(
        file_out,
        np.reshape(np.flip(rast, axis=0), (rast.size, 1)),
        fmt="%.12G",
    )

    res = dict(
        x0=x[0], y0=y[0], dx=x[1] - x[0], dy=y[1] - y[0], nx=len(x), ny=len(y)
    )

    return res


def write_job_files(
    dirs,
    param_files,
    file_job,
    job_name,
    max_time_hours=24,
    ncores_per_node=6,
    partitions="cpuall,data,datanew",
    shaltop_file="shaltop",
    folder_conf_in_job=None,
    replace_path=None,
    number_conf_file=True,
):
    """
    Write job/conf files for slurm jobs. The conf contains all the commands
    needed to run each simulation (one command per simulation).

    Parameters
    ----------
    dirs : list of string
        list of paths where simus will be run.
    param_files : list string
        list of shaltop parameter files.
    file_job : string
        name of job file called by sbatch.
    job_name : string
        name of conf file used by file_job.
    max_time_hours : int, optional
        Maximum job duration in hours before stop. The default is 24.
    ncores_per_node : int, optional
        Number of cores per nodes. Used to know the number of nodes required
        for the job. The default is 6.
    partitions : string, optional
        Names of partitions on which jobs can be launched.
        The default is "cpuall,data,datanew".
    shaltop_file : string, optional
        Bash command used to call shaltop. Can be a path.
        The default is "shaltop".
    folder_conf_in_job : string, optional
        Folder where the conf file is located. The default is the folder
        path of file_job.
    replace_path : list, optional
        replace replace_path[0] by replace_path[1] for every path in dir. This
        is used if simulations are prepared and run on two different machines
        (e.g. laptop and cluster).
        The default is None.
    number_conf_file : bool, optional
        If True, add a number in front of each line of the conf file. Required
        to identify slurm jobs.
        The default is True.

    Returns
    -------
    None.

    """
    ntasks = len(dirs)
    nnodes = int(np.ceil(ntasks / ncores_per_node))

    if folder_conf_in_job is None:
        folder_conf_in_job = os.path.dirname(file_job)
        if folder_conf_in_job == "":
            folder_conf_in_job = "."

    with open(file_job + ".conf", "w", newline="\n") as conf_file:
        if number_conf_file:
            line = "{:d} {:s} {:s} {:s}\n"
        else:
            line = "{:s} {:s} {:s}\n"
        for i in range(ntasks):
            if replace_path is not None:
                folder = dirs[i].replace(replace_path[0], replace_path[1])
                param_file = param_files[i].replace(
                    replace_path[0], replace_path[1]
                )
            else:
                folder = dirs[i]
                param_file = param_files[i]
            folder = format_path_linux(folder)
            param_file = format_path_linux(param_file)
            if number_conf_file:
                line2 = line.format(i, shaltop_file, folder, param_file)
            else:
                line2 = line.format(shaltop_file, folder, param_file)
            conf_file.write(line2)

    n_hours = np.floor(max_time_hours)
    n_min = (max_time_hours - n_hours) * 60
    str_time = "{:02.0f}:{:02.0f}:00\n".format(n_hours, n_min)

    basename = os.path.basename(file_job)
    path_conf_in_job = posixpath.join(folder_conf_in_job, basename + ".conf")

    with open(file_job + ".job", "w", newline="\n") as job_file:
        job_file.write("#!/bin/sh\n")
        job_file.write("#SBATCH -J multijob\n")
        job_file.write("#SBATCH --job-name={:s}\n".format(job_name))
        job_file.write("#SBATCH --output={:s}%j.out\n".format(job_name))
        job_file.write("#SBATCH --partition " + partitions + "\n")
        job_file.write("#SBATCH --nodes={:d}".format(nnodes) + "\n")
        job_file.write("#SBATCH --ntasks={:d}".format(ntasks) + "\n")
        job_file.write("#SBATCH --time={:s}\n".format(str_time))
        job_file.write("\n")
        job_file.write("module purge\n")
        job_file.write("module load slurm\n")
        job_file.write("\n")
        line = "srun -n {:d} -l --multi-prog {:s}"
        job_file.write(line.format(ntasks, path_conf_in_job))


def make_simus(law, rheol_params, folder_data, folder_out, readme_file):
    """
    Write shaltop initial file for simple slope test case

    Parameters
    ----------
    deltas : TYPE
        DESCRIPTION.
    folder_in : TYPE
        DESCRIPTION.
    folder_out : TYPE
        DESCRIPTION.

    Returns
    -------
    None.

    """
    # Get topography and initial mass, and write them in Shaltop format
    zfile = os.path.join(folder_data, "topo.asc")
    mfile = os.path.join(folder_data, "mass.asc")
    x, y, z, dx = tilupy.raster.read_ascii(zfile)
    _, _, m, _ = tilupy.raster.read_ascii(mfile)
    np.savetxt(os.path.join(folder_out, "z.d"), z.T.flatten())
    np.savetxt(os.path.join(folder_out, "m.d"), m.T.flatten())

    # Get simulation parameters from README.txt and raster .asc files
    params = tilupy.notations.readme_to_params(readme_file, README_PARAM_MATCH)
    params["nx"] = len(x)
    params["ny"] = len(y)
    params["per"] = dx * len(x)
    params["pery"] = dx * len(y)
    params["file_m_init"] = "../m.d"
    params["file_z_init"] = "../z.d"

    # Folder for rheological law, and set params accordingly
    folder_law = os.path.join(folder_out, law)
    params["icomp"] = SHALTOP_LAW_ID[law]

    param_names = [param for param in rheol_params]

    texts = tilupy.notations.make_rheol_string(rheol_params, law)

    # Run shaltop file
    run_shaltop_file = os.path.join(folder_law, "run_shaltop.sh")
    file_txt = ""

    for i in range(len(rheol_params[param_names[0]])):
        simu_text = texts[i]
        for param_name in param_names:
            params[param_name] = rheol_params[param_name][i]
        params["folder_output"] = simu_text
        folder_results = os.path.join(folder_law, simu_text)
        os.makedirs(folder_results, exist_ok=True)
        with open(os.path.join(folder_results, ".gitignore"), "w") as fid:
            fid.write("# Ignore everything in this directory")
            fid.write("*")
            fid.write("# Except this file")
            fid.write("!.gitignore")

        write_params_file(
            params, directory=folder_law, file_name=simu_text + ".txt"
        )
        file_txt += "start_time=`date +%s`\n"
        file_txt += 'shaltop "" ' + simu_text + ".txt\n"
        file_txt += "end_time=`date +%s`\n"
        file_txt += "elapsed_time=$(($end_time - $start_time))\n"
        file_txt += (
            'string_time="${start_time} ' + simu_text + ' ${elapsed_time}"\n'
        )
        file_txt += "echo ${string_time} >> simulation_duration.txt\n\n"

    with open(run_shaltop_file, "w") as fid:
        fid.write(file_txt)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun  1 15:27:36 2021

@author: peruzzetto
"""
import os

import numpy as np
import tilupy.read

# Dictionnary with results names lookup table, to match code output names
LOOKUP_NAMES = dict(h='hflow',
                    u='vflow', ux='vflowx', uy='vflowy',
                    hu='momentum', hu2='ek',
                    vol='vol', ekint='ekint')

# Classify results
STATES_OUTPUT = ['h', 'ux', 'uy']
forces = ['facc', 'fcurv', 'ffric', 'fgrav', 'fpression']
FORCES_OUTPUT = []
for axis in ['x', 'y']:
    for f in forces:
        FORCES_OUTPUT.append(f + axis)


def read_params(file):
    """
    Read simulation parameters from file.

    Parameters
    ----------
    file : str
        Parameters file.

    Returns
    -------
    params : dict
        Dictionnary with parameters

    """
    params = dict()
    if not file.endswith('.txt'):
        prefix = file.sep('/')[-1].sep('_')[0]
        file = os.path.join(file, prefix+'_files',
                            prefix+'_documentation.txt')

    with open(file, 'r') as f:
        for line in f:
            (key, val) = line.split('\t')
            try:
                params[key] = float(val)
            except ValueError:
                if val == 'TRUE':
                    params[key] = True
                if val == 'FALSE':
                    params[key] = False
                else:
                    params[key] = val

    return params


def read_ascii(file):
    """
    Read ascii grid file to numpy ndarray.

    Parameters
    ----------
    file : TYPE
        DESCRIPTION.

    Returns
    -------
    None.

    """
    dem = np.loadtxt(file, skiprows=6)
    # dem = np.flip(dem, axis=0).T
    grid = {}
    with open(file, 'r') as fid:
        for i in range(6):
            tmp = fid.readline().split()
            grid[tmp[0]] = float(tmp[1])
    try:
        x0 = grid['xllcenter']
        y0 = grid['yllcenter']
    except KeyError:
        x0 = grid['xllcorner']
        y0 = grid['yllcorner']
    nx = int(grid['ncols'])
    ny = int(grid['nrows'])
    dx = dy = grid['cellsize']
    x = np.linspace(x0, x0+(nx-1)*dx, nx)
    y = np.linspace(y0, y0+(ny-1)*dy, ny)

    return x, y, dem


def read_asciis(file_prefix, folder=None, ind=None, nodigit=False):
    """
    Read mutiple ascii grid file matching prefix.

    Parameters
    ----------
    file : TYPE
        DESCRIPTION.
    ind : TYPE, optional
        DESCRIPTION. The default is None.

    Returns
    -------
    None.

    """
    if folder is None:
        folder = os.getcwd()

    files_tmp = os.listdir(folder)
    files = []
    for ff in files_tmp:
        # test is file begins with file_prefix
        if ff.startswith(file_prefix) and ff.endswith('.asc'):
            letter = ff.split(file_prefix)[1][0]
            # After file_prefix, character must be either '.' or a digit
            if nodigit:
                if letter == '.':
                    files.append(ff)
            else:
                if letter.isdigit() or letter == '.':
                    files.append(ff)

    files.sort()

    if ind is not None:
        if ind == 'final':
            files = [files[-1]]
        elif ind == 'initial':
            files = [files[0]]
        else:
            files = [ff for ff in files if int(ff[-8:-4]) in ind]

    files = [os.path.join(folder, ff) for ff in files]

    dem = np.loadtxt(files[0], skiprows=6)
    if len(files) > 1:
        dem = dem[:, :, np.newaxis]
    for ff in files[1:]:
        _, _, dem2 = np.loadtxt(ff, skiprows=6)
        dem = np.concatenate((dem, dem2[:, :, np.newaxis]), axis=2)

    dem = np.squeeze(dem)

    return dem


class Results(tilupy.read.Results):
    """Results of shaltop simulations."""

    def __init__(self, text_rheol=None, folder_base=None, **varargs):
        """
        Init simulation results.

        Parameters
        ----------
        file_params : str
            File where simulation parameters will be read

        """
        self.code = 'ravaflow'
        self.prefix = text_rheol
        self.folder_base = folder_base
        file_params = os.path.join(folder_base,
                                   self.prefix + '_results',
                                   self.prefix + '_files',
                                   self.prefix + '_documentation.txt')
        self.params = read_params(file_params)

        self.folder_ascii = os.path.join(folder_base,
                                         self.prefix + '_results',
                                         self.prefix + '_ascii')
        self.folder_files = os.path.join(folder_base,
                                         self.prefix + '_results',
                                         self.prefix + '_files')

        self.set_axes()

        # Get time of outputs
        with open(os.path.join(self.folder_files,
                               self.prefix+'_summary.txt')) as fid:
            self.tim = []
            for line in fid:
                res = line.split()
                if len(res) > 0:
                    if res[0].isdigit():
                        self.tim.append(float(res[4]))

    def set_axes(self, x=None, y=None, **varargs):
        """Set x and y axes."""
        dd = self.params['Cell size']
        x = np.arange(self.params['Western boundary']+dd/2,
                      self.params['Eastern boundary'],
                      dd)
        y = np.arange(self.params['Southern boundary']+dd/2,
                      self.params['Northern boundary'],
                      dd)
        self.x = x
        self.y = y
        self.nx = len(x)
        self.ny = len(y)

    def set_zinit(self, zinit=None):
        """Set zinit, initial topography."""
        path_zinit = os.path.join(self.folder_ascii,
                                  self.prefix+'_elev.asc')
        self.zinit = np.loadtxt(path_zinit, skiprows=6)

    def get_temporal_output(self, name, d=None, t=None, **varargs):
        """
        Read 2D time dependent simulation results.

        Parameters
        ----------
        name : str
            Name of output.
        d : ndarray, optional
            Data to be read. If None, will be computed.
            The default is None.
        t : ndarray, optional
            Time of results snapshots (1D array), matching last dimension of d.
            If None, will be computed. The default is None.
        **varargs : TYPE
            DESCRIPTION.

        """
        # Read thicknesses or velocity components
        file_prefix = self.prefix + '_' + LOOKUP_NAMES[name]
        d = read_asciis(file_prefix, folder=self.folder_ascii)

        return tilupy.read.TemporalResults(name, d, self.tim)

    def get_static_output(self, name, stat,
                          d=None, from_file=False, **varargs):
        """
        Read 2D time dependent simulation results.

        Parameters
        ----------
        name : str
            Name of output.
        d : ndarray, optional
            Data to be read. Last dimension is for time.
            If None, will be computed.
            The default is None.
        **varargs : TYPE
            DESCRIPTION.

        """
        if stat in ['initial', 'final']:
            file_prefix = self.prefix + '_' + LOOKUP_NAMES[name]
            d = read_asciis(file_prefix, folder=self.folder_ascii, ind=stat)
        elif stat == 'max' and from_file:
            file = '{:s}_{:s}_{:s}.asc'.format(self.prefix,
                                               LOOKUP_NAMES[name],
                                               stat)
            file = os.path.join(self.folder_ascii, file)
            d = np.loadtxt(file, skiprows=6)
        else:
            data = self.get_temporal_output(name)
            d = data.get_temporal_stat(stat).d

        return tilupy.read.StaticResults(name+stat, d)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun  2 16:16:39 2021

@author: peruzzetto
"""

import matplotlib
import copy
import os

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
import seaborn as sns

BOLD_CONTOURS_INTV = [
    0.1,
    0.2,
    0.5,
    1,
    2.0,
    5,
    10,
    20,
    50,
    100,
    200,
    500,
    1000,
]
NB_THIN_CONTOURS = 10
NB_BOLD_CONTOURS = 3


def centered_map(cmap, vmin, vmax, ncolors=256):
    """
    Create centered colormap

    Parameters
    ----------
    cmap : TYPE
        DESCRIPTION.
    vmin : TYPE
        DESCRIPTION.
    vmax : TYPE
        DESCRIPTION.
    ncolors : TYPE, optional
        DESCRIPTION. The default is 256.

    Returns
    -------
    new_map : TYPE
        DESCRIPTION.

    """
    p = vmax / (vmax - vmin)
    npos = int(ncolors * p)
    method = getattr(plt.cm, cmap)

    colors1 = method(np.linspace(0.0, 1, npos * 2))
    colors2 = method(np.linspace(0.0, 1, (ncolors - npos) * 2))
    colors = np.concatenate(
        (colors2[: ncolors - npos, :], colors1[npos:, :]), axis=0
    )
    # colors[ncolors-npos-1,:]=np.ones((1,4))
    # colors[ncolors-npos,:]=np.ones((1,4))
    new_map = mcolors.LinearSegmentedColormap.from_list("my_colormap", colors)

    return new_map


def get_contour_intervals(
    zmin, zmax, nb_bold_contours=None, nb_thin_contours=None
):
    if nb_thin_contours is None:
        nb_thin_contours = NB_THIN_CONTOURS
    if nb_bold_contours is None:
        nb_bold_contours = NB_BOLD_CONTOURS

    intv = (zmax - zmin) / nb_bold_contours
    i = np.argmin(np.abs(np.array(BOLD_CONTOURS_INTV) - intv))

    bold_intv = BOLD_CONTOURS_INTV[i]
    if BOLD_CONTOURS_INTV[i] != BOLD_CONTOURS_INTV[0]:
        if bold_intv - intv > 0:
            bold_intv = BOLD_CONTOURS_INTV[i - 1]

    if nb_thin_contours is None:
        thin_intv = bold_intv / NB_THIN_CONTOURS
        if (zmax - zmin) / bold_intv > 5:
            thin_intv = thin_intv * 2
    else:
        thin_intv = bold_intv / nb_thin_contours

    return bold_intv, thin_intv


def auto_uniform_grey(
    z, vert_exag, azdeg=315, altdeg=45, dx=1, dy=1, std_threshold=0.01
):
    """
    Detect if shading must be applied to topography or not (uniform grey). The
    criterion in colors.LightSource.hillshade is the difference between min
    and max illumination, and seems to restrictive.
    :param z: DESCRIPTION
    :type z: TYPE
    :param vert_exag: DESCRIPTION
    :type vert_exag: TYPE
    :param dx: DESCRIPTION, defaults to 1
    :type dx: TYPE, optional
    :param dy: DESCRIPTION, defaults to 1
    :type dy: TYPE, optional
    :return: DESCRIPTION
    :rtype: TYPE

    """

    # Get topography normal direction
    e_dy, e_dx = np.gradient(vert_exag * z, dy, dx)
    normal = np.empty(z.shape + (3,)).view(type(z))
    normal[..., 0] = -e_dx
    normal[..., 1] = -e_dy
    normal[..., 2] = 1
    sum_sq = 0
    for i in range(normal.shape[-1]):
        sum_sq += normal[..., i, np.newaxis] ** 2
    normal /= np.sqrt(sum_sq)

    # Light source direction
    az = np.radians(90 - azdeg)
    alt = np.radians(altdeg)
    direction = np.array(
        [np.cos(az) * np.cos(alt), np.sin(az) * np.cos(alt), np.sin(alt)]
    )

    # Compute intensity (equivalent to LightSource hillshade, whithour rescaling)
    intensity = normal.dot(direction)
    std = np.nanstd(intensity)

    if std > std_threshold:
        return None
    else:
        return 0.5


def plot_topo(
    z,
    x,
    y,
    contour_step=None,
    nlevels=None,
    level_min=None,
    step_contour_bold="auto",
    contour_labels_properties=None,
    label_contour=True,
    contour_label_effect=None,
    axe=None,
    vert_exag=1,
    fraction=1,
    ndv=-9999,
    uniform_grey="auto",
    contours_prop=None,
    contours_bold_prop=None,
    figsize=None,
    interpolation=None,
    sea_level=0,
    sea_color=None,
    alpha=1,
    azdeg=315,
    altdeg=45,
    zmin=None,
    zmax=None,
):
    """
    Plot topography with hillshading.

    Parameters
    ----------
    z : TYPE
        DESCRIPTION.
    x : TYPE
        DESCRIPTION.
    y : TYPE
        DESCRIPTION.
    contour_step : TYPE, optional
        DESCRIPTION. The default is None.
    nlevels : TYPE, optional
        DESCRIPTION. The default is 25.
    contour_labels_properties : TYPE, optional
        DESCRIPTION. The default is None.
    axe : TYPE, optional
        DESCRIPTION. The default is None.
    vert_exag : TYPE, optional
        DESCRIPTION. The default is 1.
    contour_color : TYPE, optional
        DESCRIPTION. The default is 'k'.
    contour_alpha : TYPE, optional
        DESCRIPTION. The default is 1.
    figsize : TYPE, optional
        DESCRIPTION. The default is (10,10).
    interpolation : TYPE, optional
        DESCRIPTION. The default is None.
    sea_level : TYPE, optional
        DESCRIPTION. The default is 0.
    sea_color : TYPE, optional
        DESCRIPTION. The default is None.
    alpha : TYPE, optional
        DESCRIPTION. The default is 1.
    azdeg : TYPE, optional
        DESCRIPTION. The default is 315.
    altdeg : TYPE, optional
        DESCRIPTION. The default is 45.
    plot_terrain : TYPE, optional
        DESCRIPTION. The default is False.
    cmap_terrain : TYPE, optional
        DESCRIPTION. The default is 'gist_earth'.
    fraction : TYPE, optional
        DESCRIPTION. The default is 1.
    ndv : TYPE, optional
        DESCRIPTION. The default is -9999.

    Returns
    -------
    None.

    """
    dx = x[1] - x[0]
    dy = y[1] - y[0]
    im_extent = [x[0] - dx / 2, x[-1] + dx / 2, y[0] - dy / 2, y[-1] + dy / 2]
    ls = mcolors.LightSource(azdeg=azdeg, altdeg=altdeg)

    auto_bold_intv = None

    tmpz = np.copy(z)
    tmpz[tmpz == ndv] = np.nan

    if nlevels is None and contour_step is None:
        auto_bold_intv, contour_step = get_contour_intervals(
            np.nanmin(tmpz), np.nanmax(tmpz)
        )

    if level_min is None:
        if contour_step is not None:
            level_min = np.ceil(np.nanmin(tmpz) / contour_step) * contour_step
        else:
            level_min = np.nanmin(tmpz)
    if contour_step is not None:
        levels = np.arange(level_min, np.nanmax(tmpz), contour_step)
    else:
        levels = np.linspace(level_min, np.nanmax(tmpz), nlevels)

    if axe is None:
        fig, axe = plt.subplots(1, 1, figsize=figsize, layout="constrained")

    axe.set_ylabel("Y (m)")
    axe.set_xlabel("X (m)")
    axe.set_aspect("equal")

    if uniform_grey == "auto":
        uniform_grey = auto_uniform_grey(
            z,
            vert_exag,
            azdeg=azdeg,
            altdeg=altdeg,
            dx=dx,
            dy=dx,
        )

    z_shade = np.copy(z)
    z_shade[np.isnan(z_shade)] = ndv
    if uniform_grey is None:
        shaded_topo = ls.hillshade(
            z_shade, vert_exag=vert_exag, dx=dx, dy=dy, fraction=1
        )
    else:
        shaded_topo = np.ones(z.shape) * uniform_grey
    shaded_topo[z == ndv] = np.nan
    shaded_topo[np.isnan(z)] = np.nan
    axe.imshow(
        shaded_topo,
        cmap="gray",
        extent=im_extent,
        interpolation=interpolation,
        alpha=alpha,
        vmin=0,
        vmax=1,
    )

    if contours_prop is None:
        contours_prop = dict(alpha=0.5, colors="k", linewidths=0.5)
    axe.contour(
        x,
        y,
        np.flip(tmpz, axis=0),
        extent=im_extent,
        levels=levels,
        **contours_prop
    )

    if contours_bold_prop is None:
        contours_bold_prop = dict(alpha=0.8, colors="k", linewidths=0.8)

    if step_contour_bold == "auto":
        if auto_bold_intv is None:
            auto_bold_intv, _ = get_contour_intervals(
                np.nanmin(tmpz), np.nanmax(tmpz)
            )
        step_contour_bold = auto_bold_intv

    if step_contour_bold > 0:
        lmin = np.ceil(np.nanmin(tmpz) / step_contour_bold) * step_contour_bold
        if lmin < level_min:
            lmin = lmin + step_contour_bold
        levels = np.arange(lmin, np.nanmax(tmpz), step_contour_bold)
        cs = axe.contour(
            x,
            y,
            np.flip(tmpz, axis=0),
            extent=im_extent,
            levels=levels,
            **contours_bold_prop
        )
        if label_contour:
            if contour_labels_properties is None:
                contour_labels_properties = {}
            clbls = axe.clabel(cs, **contour_labels_properties)
            if contour_label_effect is not None:
                plt.setp(clbls, path_effects=contour_label_effect)

    if sea_color is not None:
        cmap_sea = mcolors.ListedColormap([sea_color])
        cmap_sea.set_under(color="w", alpha=0)
        mask_sea = (z <= sea_level) * 1
        if mask_sea.any():
            axe.imshow(
                mask_sea,
                extent=im_extent,
                cmap=cmap_sea,
                vmin=0.5,
                origin="lower",
                interpolation="none",
            )

    return axe


def plot_imshow(
    x,
    y,
    data,
    axe=None,
    figsize=None,
    cmap=None,
    minval=None,
    maxval=None,
    vmin=None,
    vmax=None,
    alpha=1,
    minval_abs=None,
    cmap_intervals=None,
    extend_cc="max",
    plot_colorbar=True,
    axecc=None,
    colorbar_kwargs=None,
    aspect=None,
):
    """
    plt.imshow data with some pre-processing

    Parameters
    ----------
    x : TYPE
        DESCRIPTION.
    y : TYPE
        DESCRIPTION.
    data : TYPE
        DESCRIPTION.
    axe : TYPE, optional
        DESCRIPTION. The default is None.
    figsize : TYPE, optional
        DESCRIPTION. The default is None.
    cmap : TYPE, optional
        DESCRIPTION. The default is None.
    minval : TYPE, optional
        DESCRIPTION. The default is None.
    maxval : TYPE, optional
        DESCRIPTION. The default is None.
    vmin : TYPE, optional
        DESCRIPTION. The default is None.
    vmax : TYPE, optional
        DESCRIPTION. The default is None.
    minval_abs : TYPE, optional
        DESCRIPTION. The default is None.
    cmap_intervals : TYPE, optional
        DESCRIPTION. The default is None.
    extend_cc : TYPE, optional
        DESCRIPTION. The default is "max".

    Returns
    -------
    None.

    """
    if axe is None:
        _, axe = plt.subplots(1, 1, figsize=figsize, layout="constrained")

    f = copy.copy(data)

    # vmin and vmax are similar to minval and maxval
    # and supplent minval and maxval if used
    if vmin is not None:
        minval = vmin

    if vmax is not None:
        maxval = vmax

    # Remove values below and above minval and maxval, depending on whether
    # cmap_intervals are given with or without extend_cc
    if cmap_intervals is not None:
        norm = matplotlib.colors.BoundaryNorm(
            cmap_intervals, 256, extend=extend_cc
        )
        if extend_cc in ["neither", "max"]:
            minval = cmap_intervals[0]
            f[f < minval] = np.nan
        elif extend_cc in ["neither", "min"]:
            maxval = cmap_intervals[-1]
            f[f > maxval] = np.nan
    else:
        norm = None
        # if maxval is not None:
        #     f[f > maxval] = np.nan
        if minval is not None:
            f[f < minval] = np.nan

    # Get min and max values
    if maxval is None:
        maxval = np.nanmax(f)
    if minval is None:
        minval = np.nanmin(f)

    if minval_abs:
        f[np.abs(f) <= minval_abs] = np.nan
    else:
        f[f == 0] = np.nan

    # Define colormap type
    if cmap is None:
        if maxval * minval >= 0:
            cmap = "hot_r"
        else:
            cmap = "seismic"
    if (maxval * minval >= 0) or np.isnan(maxval * minval):
        color_map = matplotlib.colormaps[cmap]
    else:
        color_map = centered_map(cmap, minval, maxval)

    if cmap_intervals is not None:
        norm = matplotlib.colors.BoundaryNorm(
            cmap_intervals, 256, extend=extend_cc
        )
        maxval = None
        minval = None
    else:
        norm = None

    # get map_extent
    dx = x[1] - x[0]
    dy = y[1] - y[0]
    im_extent = [x[0] - dx / 2, x[-1] + dx / 2, y[0] - dy / 2, y[-1] + dy / 2]

    # Plot data

    fim = axe.imshow(
        f,
        extent=im_extent,
        cmap=color_map,
        vmin=minval,
        vmax=maxval,
        alpha=alpha,
        interpolation="none",
        norm=norm,
        zorder=4,
        aspect=aspect,
    )

    # Plot colorbar
    if plot_colorbar:
        colorbar_kwargs = {} if colorbar_kwargs is None else colorbar_kwargs
        if cmap_intervals is not None and extend_cc is not None:
            colorbar_kwargs["extend"] = extend_cc
        axe.figure.colorbar(fim, cax=axecc, **colorbar_kwargs)

    return axe


def plot_data_on_topo(
    x,
    y,
    z,
    data,
    axe=None,
    figsize=(15 / 2.54, 15 / 2.54),
    cmap=None,
    minval=None,
    maxval=None,
    vmin=None,
    vmax=None,
    minval_abs=None,
    cmap_intervals=None,
    extend_cc="max",
    topo_kwargs=None,
    sup_plot=None,
    alpha=1,
    plot_colorbar=True,
    axecc=None,
    colorbar_kwargs=None,
    mask=None,
    alpha_mask=None,
    color_mask="k",
    xlims=None,
    ylims=None,
):
    """
    Plot array data on topo.

    Returns
    -------
    None.

    """

    # Initialize figure properties
    dx = x[1] - x[0]
    dy = y[1] - y[0]
    im_extent = [x[0] - dx / 2, x[-1] + dx / 2, y[0] - dy / 2, y[-1] + dy / 2]
    if axe is None:
        fig, axe = plt.subplots(1, 1, figsize=figsize, layout="constrained")

    axe.set_ylabel("Y (m)")
    axe.set_xlabel("X (m)")
    axe.set_aspect("equal", adjustable="box")

    # Plot topo
    topo_kwargs = {} if topo_kwargs is None else topo_kwargs

    if z is not None:
        plot_topo(z, x, y, axe=axe, **topo_kwargs)

    # Plot mask
    if mask is not None:
        cmap_mask = mcolors.ListedColormap([color_mask])
        cmap_mask.set_under(color="w", alpha=0)
        axe.imshow(
            mask.transpose(),
            extent=im_extent,
            cmap=cmap_mask,
            vmin=0.5,
            origin="lower",
            interpolation="none",
            # zorder=3,
            alpha=alpha_mask,
        )

    # Plot data
    plot_imshow(
        x,
        y,
        data,
        axe=axe,
        cmap=cmap,
        minval=minval,
        maxval=maxval,
        vmin=vmin,
        vmax=vmax,
        minval_abs=minval_abs,
        cmap_intervals=cmap_intervals,
        extend_cc=extend_cc,
        plot_colorbar=plot_colorbar,
        axecc=axecc,
        colorbar_kwargs=colorbar_kwargs,
    )

    # Adjust axes limits
    if xlims is not None:
        axe.set_xlim(xlims[0], xlims[1])

    if ylims is not None:
        axe.set_ylim(ylims[0], ylims[1])

    return axe


def plot_shotgather(x, t, data, xlabel="X (m)", ylabel=None, **kwargs):
    """
    Plot shotgather like image, with vertical axis as time and horizontal axis
    and spatial dimension. This is a simple call to plot_shotgather, but
    input data is transposed because in tilupy the last axis is time by
    convention.

    Parameters
    ----------
    x : NX-array
        spatial coordinates
    t : NT-array
        time array (assumed in seconds)
    data : TYPE
        NX*NT array of data to be plotted
    spatial_label : string, optional
        label for y-axis. The default is "X (m)"
    **kwargs : dict, optional
        parameters passed on to plot_imshow

    Returns
    -------
    axe : Axes
        Axes instance where data is plotted

    """
    if "aspect" not in kwargs:
        kwargs["aspect"] = "auto"
    axe = plot_imshow(x, t[::-1], data.T, **kwargs)
    axe.set_adjustable("box")
    if ylabel is None:
        ylabel = "Time (s)"
    axe.set_ylabel(ylabel)
    axe.set_xlabel(xlabel)

    return axe


def plot_maps(
    x,
    y,
    z,
    data,
    t,
    file_name=None,
    folder_out=None,
    figsize=None,
    dpi=None,
    fmt="png",
    sup_plt_fn=None,
    sup_plt_fn_args=None,
    **kwargs
):
    """
    Plot and save maps of simulations outputs at successive time steps

    Parameters
    ----------
    x : TYPE
        DESCRIPTION.
    y : TYPE
        DESCRIPTION.
    data : TYPE
        DESCRIPTION.
    t : TYPE
        DESCRIPTION.
    file_name : TYPE
        DESCRIPTION.
    folder_out : TYPE
        DESCRIPTION.
    dpi : TYPE, optional
        DESCRIPTION. The default is None.
    fmt : TYPE, optional
        DESCRIPTION. The default is 'png'.

    Returns
    -------
    None.

    """

    nfigs = len(t)
    if nfigs != data.shape[2]:
        raise ValueError(
            "length of t must be similar to the last dimension of data"
        )
    if folder_out is not None:
        file_path = os.path.join(folder_out, file_name + "_{:04d}." + fmt)
    title_fmt = "t = {:.2f} s"

    for i in range(nfigs):
        axe = plot_data_on_topo(
            x, y, z, data[:, :, i], axe=None, figsize=figsize, **kwargs
        )
        axe.set_title(title_fmt.format(t[i]))
        if sup_plt_fn is not None:
            if sup_plt_fn_args is None:
                sup_plt_fn_args = dict()
            sup_plt_fn(axe, **sup_plt_fn_args)
        # axe.figure.tight_layout(pad=0.1)
        if folder_out is not None:
            axe.figure.savefig(
                file_path.format(i),
                dpi=dpi,
                bbox_inches="tight",
                pad_inches=0.05,
            )


def colorbar(
    mappable, ax=None, cax=None, size="5%", pad=0.1, position="right", **kwargs
):
    """
    Create nice colorbar matching height/width of axe.

    Parameters
    ----------
    mappable : TYPE
        DESCRIPTION.
    cax : TYPE, optional
        DESCRIPTION. The default is None.
    size : TYPE, optional
        DESCRIPTION. The default is "5%".
    pad : TYPE, optional
        DESCRIPTION. The default is 0.05.
    position : TYPE, optional
        DESCRIPTION. The default is 'right'.
    extend : TYPE, optional
        DESCRIPTION. The default is 'neither'.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    if ax is None:
        ax = mappable.axes
    fig = ax.figure
    if position in ["left", "right"]:
        orientation = "vertical"
    else:
        orientation = "horizontal"

    # if cax is None:
    #     # divider = ax.get_axes_locator()
    #     # if divider is None:
    #     divider = make_axes_locatable(ax)
    #     cax = divider.append_axes(position, size=size, pad=pad)

    cc = fig.colorbar(mappable, cax=cax, orientation=orientation, **kwargs)

    if position == "top":
        cax.xaxis.tick_top()
        cax.xaxis.set_label_position("top")
    if position == "left":
        cax.yaxis.tick_left()
        cax.xaxis.set_label_position("left")
    return cc


def plot_heatmaps(
    df,
    values,
    index,
    columns,
    aggfunc="mean",
    figsize=None,
    ncols=3,
    heatmap_kws=None,
    notations=None,
    best_values=None,
    plot_best_value="point",
    text_kwargs=None,
):
    nplots = len(values)
    ncols = min(nplots, ncols)
    nrows = int(np.ceil(nplots / ncols))
    fig = plt.figure(figsize=figsize)
    axes = []

    for i in range(nplots):
        axe = fig.add_subplot(nrows, ncols, i + 1)
        axes.append(axe)
        data = df.pivot_table(
            index=index, columns=columns, values=values[i], aggfunc=aggfunc
        ).astype(float)
        if heatmap_kws is None:
            kws = dict()
        elif isinstance(heatmap_kws, dict):
            if values[i] in heatmap_kws:
                kws = heatmap_kws[values[i]]
            else:
                kws = heatmap_kws

        if "cmap" not in kws:
            minval = data.min().min()
            maxval = data.max().max()
            if minval * maxval < 0:
                val = max(np.abs(minval), maxval)
                kws["cmap"] = "seismic"
                kws["vmin"] = -val
                kws["vmax"] = val

        if "cbar_kws" not in kws:
            kws["cbar_kws"] = dict(pad=0.03)

        if notations is None:
            kws["cbar_kws"]["label"] = values[i]
        else:
            if values[i] in notations:
                kws["cbar_kws"]["label"] = notations[values[i]]
            else:
                kws["cbar_kws"]["label"] = values[i]

        sns.heatmap(data, ax=axe, **kws)

        if best_values is not None:
            best_value = best_values[values[i]]
            array = np.array(data)
            irow = np.arange(data.shape[0])

            if best_value == "min":
                ind = np.nanargmin(array, axis=1)
                i2 = np.nanargmin(array[irow, ind])
            if best_value == "min_abs":
                ind = np.nanargmin(np.abs(array), axis=1)
                i2 = np.nanargmin(np.abs(array[irow, ind]))
            elif best_value == "max":
                ind = np.nanargmax(array, axis=1)
                i2 = np.nanargmax(array[irow, ind])

            if plot_best_value == "point":
                axe.plot(
                    ind + 0.5,
                    irow + 0.5,
                    ls="",
                    marker="o",
                    mfc="w",
                    mec="k",
                    mew=0.5,
                    ms=5,
                )
                axe.plot(
                    ind[i2] + 0.5,
                    i2 + 0.5,
                    ls="",
                    marker="o",
                    mfc="w",
                    mec="k",
                    mew=0.8,
                    ms=9,
                )
            elif plot_best_value == "text":
                indx = list(ind)
                indx.pop(i2)
                indy = list(irow)
                indy.pop(i2)
                default_kwargs = dict(ha="center", va="center", fontsize=8)
                if text_kwargs is None:
                    text_kwargs = default_kwargs
                else:
                    text_kwargs = dict(default_kwargs, **text_kwargs)
                for i, j in zip(indx, indy):
                    axe.text(
                        i + 0.5,
                        j + 0.5,
                        "{:.2g}".format(array[j, i]),
                        **text_kwargs
                    )
                text_kwargs2 = dict(text_kwargs, fontweight="bold")
                axe.text(
                    ind[i2] + 0.5,
                    i2 + 0.5,
                    "{:.2g}".format(array[i2, ind[i2]]),
                    **text_kwargs2
                )

    axes = np.array(axes).reshape((nrows, ncols))
    for i in range(nrows):
        for j in range(1, ncols):
            axes[i, j].set_ylabel("")
            # axes[i, j].set_yticklabels([])

    for i in range(nrows - 1):
        for j in range(ncols):
            axes[i, j].set_xlabel("")
            # axes[i, j].set_xticklabels([])

    if notations is not None:
        for i in range(nrows):
            axes[i, 0].set_ylabel(notations[index])
        for j in range(ncols):
            axes[-1, j].set_xlabel(notations[columns])

    # fig.tight_layout()

    return fig



# -*- coding: utf-8 -*-
"""
Created on Fri Aug  4 19:27:09 2023

@author: peruzzetto
"""

import numpy as np


def calotte(x, y, z, x0, y0, radius, norm_offset=0, res_type='projected_normal'):
    """
    Construct mass on topography as volume between sphere and topography.

    Parameters
    ----------
    x : np.array
        xaxis array, with length nx
    y : np.array
        yaxis array, with length ny
    z : np.array
        array of altitudes, of size (ny, nx). z[0, 0] has coordinates
        (x[0], y[-1])
    x0 : float
        x position of the calotte
    y0 : float
        y position of the calotte
    radius : float
        radius of the shpere
    norm_offset : float, optional
        downwards offset between the sphere center and the topography, in the
        direction normal to the topography. The default is 0.
    res_type : string, optional
        Type of thickness output. 'true_normal': real thickness in the
        direction normal to the topography. 'vertical': Thickness in the
        vertical direction. 'projected_normal': Thickness normal to the
        topography is computed from the vertical thickness projected on 
        the axe normal to the topography. The default is 'projected_normal'.

    Returns
    -------
    m : np.array
        array of mass height, in the direction normal to topography

    """
    z = np.flip(z, axis=0).T

    xmesh, ymesh = np.meshgrid(x, y, indexing='ij')
    nx = len(x)
    ny = len(y)

    # Get altitude of mass center on topography
    i0 = np.unravel_index(np.argmin(np.abs(x-x0), axis=None), (nx,))
    j0 = np.unravel_index(np.argmin(np.abs(y-y0), axis=None), (ny,))
    z0 = z[i0, j0]
    # Topography gradient
    [Fx, Fy] = np.gradient(z, x, y, edge_order=2)
    Fz = np.ones((nx, ny))
    c = 1/np.sqrt(1+Fx**2+Fy**2)
    Fx = -Fx*c
    Fy = -Fy*c
    Fz = Fz*c
    # Correct position from offset (shpere is moved downward,
    # perpendicular to topography)
    x0 = x0-norm_offset*Fx[i0, j0]
    y0 = y0-norm_offset*Fy[i0, j0]
    z0 = z0-norm_offset*Fz[i0, j0]

    # Compute mass height only where relevant (ie around the mass center)
    dist_to_mass = (xmesh-x0)**2+(ymesh-y0)**2
    ind = (dist_to_mass <= radius**2)

    B = 2*(Fx*(xmesh-x0)+Fy *
           (ymesh-y0)+Fz*(z-z0))
    C = (xmesh-x0)**2+(ymesh-y0)**2+(z-z0)**2-radius**2
    D = B**2-4*C

    # Intersection between shpere and normal to the topography, solution of
    # t**2+B*t+C=0
    m = np.zeros((nx, ny))

    if res_type == 'true_normal':
        # B = 2*(Fx[ind]*(xmesh[ind]-x0)+Fy[ind] *
        #        (ymesh[ind]-y0)+Fz[ind]*(z[ind]-z0))
        # C = (xmesh[ind]-x0)**2+(ymesh[ind]-y0)**2+(z[ind]-z0)**2-radius**2
        # D = B**2-4*C
        B = 2*(Fx*(xmesh-x0)+Fy *
               (ymesh-y0)+Fz*(z-z0))
        C = (xmesh-x0)**2+(ymesh-y0)**2+(z-z0)**2-radius**2
        D = B**2-4*C
        ind = D > 0
        t1 = (-B-np.sqrt(D))/2
        t2 = (-B+np.sqrt(D))/2
        ind2 = t1*t2 < 0
        m[ind2] = np.maximum(t1[ind2], t2[ind2])

    # Vertical thickness of calotte.
    if res_type in ['vertical', 'projected_normal']:
        zs = z0 + np.sqrt(radius**2 - (xmesh - x0)**2 - (ymesh - y0)**2)
        zi = z0 - np.sqrt(radius**2 - (xmesh - x0)**2 - (ymesh - y0)**2)
        ind = (z < zs) & (z > zi)
        m[ind] = zs[ind] - z[ind]
        if res_type == 'projected_normal':
            m = m * c

    m = np.flip(m.T, axis=0)

    return m



# -*- coding: utf-8 -*-
"""
Created on Fri May 31 11:23:25 2024

@author: peruzzetto
"""

import warnings
import numpy as np
import os

from scipy.interpolate import RegularGridInterpolator

import tilupy.raster


def make_edges_matrices(nx, ny):
    """
    Numbering edges for a regular rectangular grid.

    Considering a matrix M with whape (ny, nx),the convention is that
    M[0, 0] corresponds to the lower left corner of the matrix, M[0, -1]
    to the lower right corner, M[-1, 0] to the upper left and M[-1, -1] to
    the upper right. Edges are numbered cell by cell, counter clockwise :
    bottom, right, up, left. The first cell is M[0, 0], then cells are
    processed line by line.

    :param nx: Number of faces in the X direction
    :type nx: int
    :param ny: Number of faces in the Y direction
    :type ny: int
    :return: DESCRIPTION
    :rtype: TYPE

    """

    # Numbers of horizontal edges
    h_edges = np.zeros((ny + 1, nx))
    # Numbers of vertical edges
    v_edges = np.zeros((ny, nx + 1))

    # Number of edges on first completed line
    n_edges_l1 = 4 + 3 * (nx - 1)
    # Number of edges on folowwing lines
    n_edges_l = 3 + 2 * (nx - 1)

    # Fill first line of h_edges
    h_edges[0, 0] = 1
    h_edges[0, 1:] = np.arange(nx - 1) * 3 + 5
    # Fill second line of h_edges
    h_edges[1, :] = h_edges[0, :] + 2

    try:
        # Fill first column of h_edges
        h_edges[2:, 0] = np.arange(ny - 1) * n_edges_l + n_edges_l1 + 2
        # Fill the rest of the table
        tmp = np.concatenate(([3], np.arange(nx - 2) * 2 + 5))
        h_edges[2:, 1:] = h_edges[2:, 0][:, np.newaxis] + tmp[np.newaxis, :]
    except IndexError:
        pass

    # Fill first  2 columns of v_edges
    v_edges[0, 0] = 4
    try:
        v_edges[1:, 0] = np.arange(ny - 1) * n_edges_l + n_edges_l1 + 3
    except IndexError:
        pass
    v_edges[:, 1] = v_edges[:, 0] - 2
    # Fill the rest of v_edges
    try:
        v_edges[0, 2:] = 2 + np.concatenate(([4], np.arange(nx - 2) * 3 + 7))
        tmp = np.concatenate(([3], np.arange(nx - 2) * 2 + 5))
        v_edges[1:, 2:] = v_edges[1:, 1][:, np.newaxis] + tmp[np.newaxis, :]
    except IndexError:
        pass

    h_edges = np.flip(h_edges, axis=0).astype(int)
    v_edges = np.flip(v_edges, axis=0).astype(int)
    return h_edges, v_edges


if __name__ == "__main__":
    h_edges, v_edges = make_edges_matrices(1, 1)


class ModellingDomain:
    """
    Mesh of simulation topography
    """

    def __init__(
        self,
        raster=None,
        xmin=0,
        ymin=0,
        nx=None,
        ny=None,
        dx=1,
        dy=1,
    ):
        if raster is not None:
            if isinstance(raster, np.ndarray):
                self.z = raster
                self.nx = raster.shape[1]
                self.ny = raster.shape[0]
                self.dx = dx
                self.dy = dy
                self.x = np.arange(
                    xmin, xmin + (self.nx - 1) * self.dx + self.dx / 2, self.dx
                )
                self.y = np.arange(
                    ymin, ymin + (self.ny - 1) * self.dy + self.dy / 2, self.dy
                )
            if isinstance(raster, str):
                self.x, self.y, self.z = tilupy.raster.read_raster(raster)
                self.nx = len(self.x)
                self.ny = len(self.y)
                self.dx = self.x[1] - self.x[0]
                self.dy = self.y[1] - self.y[0]
        else:
            self.z = raster
            self.nx = nx
            self.ny = ny
            self.dx = dx
            self.dy = dy
            if (
                xmin is not None
                and ymin is not None
                and nx is not None
                and ny is not None
                and dx is not None
                and dy is not None
            ):
                self.x = np.arange(xmin, xmin + dx * nx + dx / 2, dx)
                self.y = np.arange(ymin, ymin + dy * ny + dy / 2, dy)

        self.h_edges = None
        self.v_edges = None

        if self.nx is not None and self.ny is not None:
            self.set_edges()

    def set_edges(self):
        """
        Set horizontal and vertical edges number
        :return: DESCRIPTION
        :rtype: TYPE

        """
        self.h_edges, self.v_edges = make_edges_matrices(
            self.nx - 1, self.ny - 1
        )

    def get_edge(self, xcoord, ycoord, cardinal):
        """
        Get edge number for given coordinates and cardinal direction
        :param xcoord: DESCRIPTION
        :type xcoord: TYPE
        :param ycoord: DESCRIPTION
        :type ycoord: TYPE
        :param cardinal: DESCRIPTION
        :type cardinal: TYPE
        :return: DESCRIPTION
        :rtype: TYPE

        """
        assert (self.h_edges is not None) and (
            self.v_edges is not None
        ), "h_edges and v_edges must be computed to get edge number"
        ix = np.argmin(np.abs(self.x[:-1] + self.dx / 2 - xcoord))
        iy = np.argmin(np.abs(self.y[:-1] + self.dx / 2 - ycoord))
        if cardinal == "S":
            return self.h_edges[-iy - 1, ix]
        elif cardinal == "N":
            return self.h_edges[-iy - 2, ix]
        elif cardinal == "W":
            return self.v_edges[-iy - 1, ix]
        elif cardinal == "E":
            return self.v_edges[-iy - 1, ix + 1]

    def get_edges(self, xcoords, ycoords, cardinal, from_extremities=True):
        """
        Get edges numbers for a set of coordinates and cardinals direction.
        xcoords and ycoords can be the min and max x and y coordinates of a line
        if from_extremities is True.
        :param xcoords: DESCRIPTION
        :type xcoords: TYPE
        :param ycoords: DESCRIPTION
        :type ycoords: TYPE
        :param cardinal: DESCRIPTION
        :type cardinal: TYPE
        :param from_extremities: DESCRIPTION, defaults to True
        :type from_extremities: TYPE, optional
        :return: DESCRIPTION
        :rtype: TYPE

        """
        if from_extremities:
            d = np.sqrt(
                (xcoords[1] - xcoords[0]) ** 2 + (ycoords[1] - ycoords[0]) ** 2
            )
            npts = int(np.ceil(d / min(self.dx, self.dy)))
            xcoords = np.linspace(xcoords[0], xcoords[1], npts + 1)
            ycoords = np.linspace(ycoords[0], ycoords[1], npts + 1)
        if self.h_edges is None or self.v_edges is None:
            self.set_edges
        res = dict()
        # cardinal is a string with any combination of cardinal directions
        for card in cardinal:
            edges_num = []
            for xcoord, ycoord in zip(xcoords, ycoords):
                edges_num.append(self.get_edge(xcoord, ycoord, card))
            res[card] = list(set(edges_num))  # Duplicate edges are removed
            res[card].sort()
        return res


class Simu:
    def __init__(self, folder, name):
        os.makedirs(folder, exist_ok=True)
        self.folder = folder
        self.name = name
        self.x = None
        self.y = None
        self.z = None

    def set_topography(self, z, x=None, y=None, file_out=None):
        if isinstance(z, str):
            self.x, self.y, self.z = tilupy.raster.read_raster(z)
        else:
            self.x, self.y, self.z = x, y, z

        if file_out is None:
            file_out = "toposimu.asc"
        else:
            fname = file_out.split(".")[0]
            if len(fname) > 8:
                warnings.warns(
                    """Topography file name is too long,
                    only 8 first characters are retained"""
                )
                fname = fname[:8]
            elif len(fname) < 8:
                warnings.warns(
                    """Topography file name is too short and is adapted,
                    exactly 8 characters needed"""
                )
                fname = fname.ljust(8, "0")
            file_out = fname + ".asc"

        tilupy.raster.write_ascii(
            self.x, self.y, self.z, os.path.join(self.folder, file_out)
        )

    def set_numeric_params(
        self,
        tmax,
        dtsorties,
        paray=0.00099,
        dtinit=0.01,
        cfl_const=1,
        CFL=0.2,
        alpha_cor_pente=1.0,
    ):
        self.tmax = tmax
        self.dtsorties = dtsorties

        with open(os.path.join(self.folder, "DONLEDD1.DON"), "w") as fid:
            fid.write(
                "paray".ljust(34, " ") + "{:.12f}\n".format(paray).lstrip("0")
            )
            fid.write("tmax".ljust(34, " ") + "{:.12f}\n".format(tmax))
            fid.write("dtinit".ljust(34, " ") + "{:.12f}\n".format(dtinit))
            fid.write(
                "cfl constante ?".ljust(34, " ") + "{:.0f}\n".format(cfl_const)
            )
            fid.write("CFL".ljust(34, " ") + "{:.12f}\n".format(CFL))
            fid.write("Go,VaLe1,VaLe2".ljust(34, " ") + "{:.0f}\n".format(3))
            fid.write("secmbr0/1".ljust(34, " ") + "{:.0f}\n".format(1))
            fid.write(
                "CL variable (si=0) ?".ljust(34, " ") + "{:d}\n".format(0)
            )
            fid.write(
                "dtsorties".ljust(34, " ") + "{:.12f}\n".format(dtsorties)
            )
            fid.write(
                "alpha cor pentes".ljust(34, " ")
                + "{:.12f}".format(alpha_cor_pente)
            )

    def set_rheology(self, tau_rho, K_tau=0.3):
        """
        Set Herschel-Bulkley rheology parameters

        Rheological parameters are tau/rho and K/tau. See
        -Coussot, P., 1994. Steady, laminar, flow of concentrated mud
        suspensions in open channel. Journal of Hydraulic Research 32, 535–559.
        doi.org/10.1080/00221686.1994.9728354 --> Eq 8 and text after eq 22
        for the default value of K/tau
        -Rickenmann, D. et al., 2006. Comparison of 2D debris-flow
        simulation models with field events. Computational Geosciences 10,
        241–264. doi.org/10.1007/s10596-005-9021-3 --> Eq 9


        :param tau_rho: DESCRIPTION
        :type tau_rho: TYPE
        :param K_rau: DESCRIPTION
        :type K_rau: TYPE
        :return: DESCRIPTION
        :rtype: TYPE

        """
        with open(os.path.join(self.folder, self.name + ".rhe"), "w") as fid:
            fid.write("{:.3f}\n".format(tau_rho))
            fid.write("{:.3f}".format(K_tau))

    def set_boundary_conditions(
        self,
        xcoords,
        ycoords,
        cardinal,
        discharges,
        times=None,
        thicknesses=None,
        tmax=9999,
        discharge_from_volume=False,
    ):
        try:
            modelling_domain = ModellingDomain(
                self.z,
                self.x[0],
                self.y[0],
                dx=self.x[1] - self.x[0],
                dy=self.y[1] - self.y[0],
            )
        except AttributeError:
            print("Simulation topography has not been set")
            raise

        edges = modelling_domain.get_edges(
            xcoords, ycoords, cardinal, from_extremities=True
        )
        edges = edges[cardinal]

        if times is None:
            # If no time vector is given, discharges is interpreted as a volume
            # from which an empirical peak discharge is computed
            peak_discharge = 0.0188 * discharges**0.79
            times = [0, 2 * discharges / peak_discharge, tmax]
            discharges = [peak_discharge, 0, 0]

        n_times = len(times)
        n_edges = len(edges)

        assert n_times == len(
            discharges
        ), "discharges and time vectors must be the same length"

        if thicknesses is None:
            thicknesses = [1 for i in range(n_times)]

        if cardinal in ["W", "E"]:
            dd = self.y[1] - self.y[0]
        else:
            dd = self.x[1] - self.x[0]

        with open(os.path.join(self.folder, self.name + ".cli"), "w") as fid:
            fid.write(
                "{:d}\n".format(n_times)
            )  # Number of time steps in the hydrogram
            for i, time in enumerate(times):
                fid.write("{:.4f}\n".format(time))  # Write time
                fid.write("{:d}\n".format(n_edges))  # Write n_edges
                qn = discharges[i] / (n_edges * dd)
                qt = 0
                for i_edge, edge in enumerate(edges):
                    fid.write(
                        "\t{:d} {:.7E} {:.7E} {:.7E}\n".format(
                            edge, thicknesses[i], qn, qt
                        )
                    )

    def set_init_mass(self, mass_raster, h_min=0.01):
        x, y, m = tilupy.raster.read_ascii(mass_raster)
        # The input raster is given as the topography for the grid corners,
        # but results are then written on grid cell centers
        x2 = x[1:] - (x[1] - x[0]) / 2
        y2 = y[1:] - (y[1] - y[0]) / 2
        fm = RegularGridInterpolator(
            (y, x),
            m,
            method="linear",
            bounds_error=False,
            fill_value=None,
        )
        x_mesh, y_mesh = np.meshgrid(x2, y2)
        m2 = fm((y_mesh, x_mesh))
        m2[m2 < h_min] = h_min
        np.savetxt(
            os.path.join(self.folder, self.name + ".cin"),
            m2.flatten(),
            header="0.000000E+00",
            comments="",
            fmt="%.10E",
        )


if __name__ == "__main__":
    domain = ModellingDomain(xmin=0, ymin=0, nx=4, ny=3, dx=1, dy=1)
    domain.set_edges()
    res = domain.get_edges([1.5, 2.5], [0, 0], "S")



# -*- coding: utf-8 -*-
"""
Created on Thu Jul 13 10:39:05 2023

@author: peruzzetto
"""

import requests

import os

def import_frankslide_dem(folder_out=None, file_out=None):
    
    if folder_out is None:
        folder_out = '.'
    if file_out is None:
        file_out = 'Frankslide_topography.asc'
    
    file_save = os.path.join(folder_out, file_out)
    
    url = ('https://raw.githubusercontent.com/marcperuz/tilupy/main/data/'+
           'frankslide/rasters/Frankslide_topography.asc')
    r = requests.get(url)
    open(file_save, 'w').write(r.text)
    
    return file_save

def import_frankslide_pile(folder_out=None, file_out=None):
    
    if folder_out is None:
        folder_out = '.'
    if file_out is None:
        file_out = 'Frankslide_pile.asc'
    
    file_save = os.path.join(folder_out, file_out)
    
    url = ('https://raw.githubusercontent.com/marcperuz/tilupy/main/data/'+
           'frankslide/rasters/Frankslide_pile.asc')
    r = requests.get(url)
    open(file_save, 'w').write(r.text)
    
    return file_save


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jun  3 12:20:34 2021

@author: peruzzetto
"""
import pandas as pd

LABEL_OPTIONS = dict(language="english", label_type="symbol")


class Notation:
    def __init__(
        self, name, long_name=None, gender=None, symbol=None, unit=None
    ):
        self.name = name
        self.long_name = long_name
        self.gender = gender
        self.symbol = symbol
        self.unit = unit

    @property
    def unit(self):
        return self._unit

    @unit.setter
    def unit(self, value):
        if value is None:
            self._unit = None
        else:
            self._unit = value

    @property
    def gender(self):
        return self._gender

    @gender.setter
    def gender(self, value):
        if value is None:
            self._gender = Gender()
        elif isinstance(value, dict):
            self._gender = Gender(**value)
        else:
            self._gender = value

    @property
    def long_name(self):
        return self._long_name

    @long_name.setter
    def long_name(self, value):
        if value is None:
            self._long_name = LongName()
        elif isinstance(value, dict):
            self._long_name = LongName(**value)
        else:
            self._long_name = value

    def get_long_name(self, language=None, gender=None):
        if isinstance(self.long_name, str):
            return self.long_name

        if language is None:
            language = LABEL_OPTIONS["language"]

        res = getattr(self.long_name, language)
        if gender is not None:
            res = res[gender]

        return res


class Unit(pd.Series):
    UNITS = ["Pa", "N", "kg", "m", "s"]

    def __init__(self, series=None, **kwargs):
        if series is not None:
            super().__init__(series)
        else:
            super().__init__()
            for key in kwargs:
                if key not in Unit.UNITS:
                    raise ValueError("unrecognized unit")
                self[key] = kwargs[key]

    def __mul__(self, other):
        tmp = self.add(other, fill_value=0)
        return Unit(tmp[tmp != 0])

    def get_label(self):
        if self.empty:
            return ""

        positives = self[self >= 1].reindex(Unit.UNITS).dropna()
        negatives = self[self < 0].reindex(Unit.UNITS).dropna()
        text_label = [
            index + "$^{:.0f}$".format(positives[index])
            for index in positives.index
        ]
        text_label += [
            index + "$^{{{:.0f}}}$".format(negatives[index])
            for index in negatives.index
        ]
        text_label = " ".join(text_label)
        text_label = text_label.replace("$^1$", "")

        return text_label


class LongName(object):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class Gender(object):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


NOTATIONS = dict()

NOTATIONS["x"] = Notation(
    "x",
    symbol="X",
    unit=Unit(m=1),
    long_name=LongName(english="X", french="X"),
    gender=Gender(english=None, french="f"),
)
NOTATIONS["y"] = Notation(
    "y",
    symbol="Y",
    unit=Unit(m=1),
    long_name=LongName(english="Y", french="Y"),
    gender=Gender(english=None, french="f"),
)
NOTATIONS["t"] = Notation(
    "t",
    symbol="t",
    unit=Unit(s=1),
    long_name=LongName(english="Time", french="Temps"),
    gender=Gender(english=None, french="m"),
)
NOTATIONS["xy"] = Notation(
    "xy",
    symbol="XY",
    unit=Unit(m=2),
    long_name=LongName(english="Time", french="Temps"),
    gender=Gender(english=None, french="m"),
)
NOTATIONS["h"] = Notation(
    "h",
    symbol="h",
    unit=Unit(m=1),
    long_name=LongName(english="thickness", french="épaisseur"),
    gender=Gender(english=None, french="f"),
)
NOTATIONS["hvert"] = Notation(
    "hvert",
    symbol="h^v",
    unit=Unit(m=1),
    long_name=LongName(
        english="vertical thickness", french="épaisseur verticale"
    ),
    gender=Gender(english=None, french="f"),
)
NOTATIONS["u"] = Notation(
    "u",
    symbol="u",
    unit=Unit(m=1, s=-1),
    long_name=LongName(english="velocity", french="vitesse"),
    gender=Gender(english=None, french="f"),
)
NOTATIONS["hu"] = Notation(
    "hu",
    symbol="hu",
    unit=Unit(m=2, s=-1),
    long_name=None,
    gender=None,
)

NOTATIONS["max"] = Notation(
    "max",
    symbol="max",
    unit=None,
    long_name=LongName(
        english="maximum", french=dict(m="maximum", f="maximale")
    ),
    gender=None,
)
NOTATIONS["int"] = Notation(
    "int",
    symbol="int",
    unit=None,
    long_name=LongName(
        english="maximum", french=dict(m="maximum", f="maximale")
    ),
    gender=None,
)


def get_notation(name, language=None):
    try:
        notation = NOTATIONS[name]
    except KeyError:
        strings = name.split("_")
        if len(strings) == 1:
            notation = Notation(
                name, symbol=name, unit=None, long_name=name, gender=None
            )
        else:
            state = get_notation(strings[0])
            operator = get_notation(strings[1])
            if len(strings) == 3:
                axis = strings[2]
            else:
                axis = None
            notation = add_operator(
                state, operator, axis=axis, language=language
            )

    return notation


def get_operator_unit(name, axis):
    if name == "int":
        if axis == "t" or axis is None:
            unit = Unit(s=1)
        if axis in ["x", "y"]:
            unit = Unit(m=1)
        if axis == "xy":
            unit = Unit(m=2)
    else:
        unit = Unit(pd.Series())
    return unit


def make_long_name(notation, operator, language=None):
    if language is None:
        language = LABEL_OPTIONS["language"]

    str_notation = notation.get_long_name(language=language)
    try:
        gender = gender = getattr(notation.gender, language)
    except AttributeError:
        gender = None
    str_operator = operator.get_long_name(language=language, gender=gender)

    if language == "english":
        res = str_operator + " " + str_notation
    elif language == "french":
        res = str_notation + " " + str_operator

    return res


def add_operator(notation, operator, axis=None, language=None):
    if isinstance(operator, str):
        operator = get_notation(operator)

    if isinstance(notation, str):
        notation = get_notation(notation)

    operator_symbol = operator.symbol
    if axis is not None:
        operator_symbol += "({})".format(axis)

    unit_operator = get_operator_unit(operator.name, axis)

    if operator.name == "int":
        if axis is None:
            ll = "t"  # If axis is not specified by default integration is over time
        else:
            ll = axis
        symbol = "\\int_{{{}}} {}".format(ll, notation.symbol)
    else:
        operator_symbol = operator.symbol
        if axis is not None:
            operator_symbol += "({})".format(axis)
        symbol = notation.symbol + "_{{{}}}".format(operator_symbol)

    if notation.unit is None:
        unit = None
    else:
        unit = notation.unit * unit_operator

    res = Notation(
        name=notation.name + "_" + operator.name,
        symbol=symbol,
        unit=unit,
        long_name=make_long_name(notation, operator, language=language),
    )
    return res


def get_label(notation, with_unit=True, label_type=None, language=None):
    if isinstance(notation, str):
        notation = get_notation(notation)

    if label_type is None:
        label_type = LABEL_OPTIONS["label_type"]
    if language is None:
        language = LABEL_OPTIONS["language"]

    if label_type == "litteral":
        label = notation.get_long_name(language=language, gender=None)
    elif label_type == "symbol":
        label = "$" + notation.symbol + "$"

    if with_unit and notation.unit is not None:
        unit_string = notation.unit.get_label()
        # Add unit only if string is not empty
        if unit_string:
            label = label + " ({})".format(unit_string)

    return label


def set_label_options(**kwargs):
    global LABEL_OPTIONS
    LABEL_OPTIONS.update(**kwargs)


def readme_to_params(file, readme_param_match=None):
    """
    Convert README file as dictionnary of parameters, following matching names.

    Parameters
    ----------
    file : TYPE
        DESCRIPTION.
    readme_param_match : TYPE
        DESCRIPTION.

    Returns
    -------
    params : TYPE
        DESCRIPTION.

    """
    params = dict()
    with open(file, "r") as f:
        if readme_param_match is not None:
            for line in f:
                (key, val) = line.split()
                if key in readme_param_match:
                    params[readme_param_match[key]] = val
        else:
            for line in f:
                (key, val) = line.split()
                params[key] = val

    return params


def make_rheol_string_fmt(rheoldict, law="coulomb"):
    """Make string from rheological parameters."""
    text = ""
    for name in ["delta1", "delta2", "delta3", "delta4"]:
        if name in rheoldict:
            new_txt = name + "_{:05.2f}_"
            text += new_txt
    if "ksi" in rheoldict:
        new_txt = name + "ksi_{:06.1f}_"
        text += new_txt
    text = text[:-1]

    return text


def make_rheol_string(rheoldict, law):
    """Make strings from rheological parameters."""

    keys = [key for key in rheoldict]
    for key in keys:
        if not isinstance(rheoldict[key], list):
            rheoldict[key] = [rheoldict[key]]

    nparams = len(rheoldict[keys[0]])
    txt_fmt = make_rheol_string_fmt(rheoldict)
    texts = []

    for i in range(nparams):
        if law == "coulomb":
            txt_fmt = "delta1_{:05.2f}"
            text = txt_fmt.format(rheoldict["delta1"][i]).replace(".", "p")
        if law == "voellmy":
            txt_fmt = "delta1_{:05.2f}_ksi_{:06.1f}"
            text = txt_fmt.format(
                rheoldict["delta1"][i], rheoldict["ksi"][i]
            ).replace(".", "p")
        if law == "pouliquen_2002":
            txt_fmt = "delta1_{:05.2f}_L_{:05.2f}"
            text = txt_fmt.format(
                rheoldict["delta1"][i], rheoldict["wlong"][i]
            ).replace(".", "p")
        texts.append(text)

    if nparams == 1:
        texts = texts[0]

    return texts



# -*- coding: utf-8 -*-
"""
Created on Mon Oct 10 17:31:10 2022

@author: peruzzetto
"""

import numpy as np
import pandas as pd

from tilupy import read
from tilupy import utils
from tilupy import raster


def CSI(simu, observation=None, h_threshs=[1], state='h_final'):

    res = []
    assert(observation is not None)

    if isinstance(observation, str):
        _, _, observation, _ = raster.read_raster(observation)

    strs = state.split('_')
    if state == 'h_max':
        d = simu.h_max
    else:
        d = simu.get_static_output(strs[0], strs[1]).d
    for h_thresh in h_threshs:
        array = 1*d > h_thresh
        res.append(utils.CSI(array, observation))

    return h_threshs, res


def diff_runout(simu, point=None, h_threshs=[1],
                section=None, orientation='W-E',
                state='h_max',
                get_contour_kws=None):

    res = []
    assert(point is not None)

    if get_contour_kws is None:
        get_contour_kws = dict()

    if state == 'h_max':
        d = simu.h_max
    else:
        strs = state.split('_')
        d = simu.get_static_output(strs[0], strs[1]).d
    xc, yc = utils.get_contour(simu.x, simu.y, d, h_threshs, **get_contour_kws)

    for h in h_threshs:
        res.append(utils.diff_runout(xc[h], yc[h], point, section=section,
                                     orientation=orientation))

    return h_threshs, res


def eval_simus(simus, methods, calib_parameters, methods_kws,
               code='shaltop',
               recorded_params=['delta1'],
               calib_parameter_name='h_threshs'):
    """
    Evaluate simulation results with different methods

    Parameters
    ----------
    simus : TYPE
        DESCRIPTION.
    methods : TYPE
        DESCRIPTION.

    Returns
    -------
    None.

    """
    if not isinstance(methods, list):
        methods = [methods]
    if not isinstance(methods_kws, list):
        methods_kws = [methods_kws]

    if isinstance(simus, dict):
        simus = pd.DataFrame(simus)
    if isinstance(simus, pd.DataFrame):
        simus_list = []
        for i in range(simus.shape[0]):
            simus_list.append(read.get_results(
                code, **simus.iloc[i, :].to_dict()))
        simus2 = simus.copy()
    else:
        simus_list = simus
        simus2 = pd.DataFrame()

    for param in recorded_params:
        simus2[param] = np.nan
    fn = dict()
    for method in methods:
        fn[method] = globals()[method]
        simus2[method] = np.nan
    simus2[calib_parameter_name] = np.nan

    ns = len(simus_list)
    nc = len(calib_parameters)
    nn = ns*nc

    res = pd.DataFrame(columns=simus2.columns, index=np.arange(nn))

    for i, simu in enumerate(simus_list):
        # Initiate fields
        istart = i*nc
        iend = (i+1)*nc
        res.iloc[istart:iend, :] = simus2.loc[i, :].copy()
        for j, method in enumerate(methods):
            kws = methods_kws[j]
            kws[calib_parameter_name] = calib_parameters
            _, calib_res = fn[method](simu, **kws)
            for param in recorded_params:
                res.loc[:, param].iloc[istart:iend] = simu.params[param]
            res.loc[:, calib_parameter_name].iloc[istart:iend] = calib_parameters
            res.loc[:, method].iloc[istart:iend] = calib_res

    return res
